###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Moving toward a templates-less system."""

import ast

from DIRAC.Core.Base.Script import Script

Script.parseCommandLine()

from DIRAC import gConfig, gLogger, exit as DIRACexit
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest


def fillVisList(vdict, num):
    # Assuming that, if there's only one element in the list of output visibility flags, every step will catch that flag
    if len(vdict) == 1:
        val = vdict[list(vdict)[0]]
        vdict = {str(i): val for i in range(int(list(vdict)[0]), int(list(vdict)[0]) + num)}
    # Another assumption: if the number of steps is bigger than that of vis flags,
    # then extend the list with the last flag available
    # to fill the "holes"
    # if len(vlist) < len(slist):
    #  vlist.extend( vlist[-1] * (len(slist) - len(vlist)) )

    return vdict


gLogger = gLogger.getSubLogger("LaunchingRequest_run.py")
currentSetup = gConfig.getValue("DIRAC/Setup")

pr = ProductionRequest()

stepsList = ["{{p1Step}}"]
stepsList.append("{{p2Step}}")
stepsList.append("{{p3Step}}")
stepsList.append("{{p4Step}}")
stepsList.append("{{p5Step}}")
stepsList.append("{{p6Step}}")
stepsList.append("{{p7Step}}")
stepsList.append("{{p8Step}}")
stepsList.append("{{p9Step}}")
stepsList.append("{{p10Step}}")
stepsList.append("{{p11Step}}")
stepsList.append("{{p12Step}}")
stepsList.append("{{p13Step}}")
stepsList.append("{{p14Step}}")
stepsList.append("{{p15Step}}")
stepsList.append("{{p16Step}}")
stepsList.append("{{p17Step}}")
stepsList.append("{{p18Step}}")
stepsList.append("{{p19Step}}")
stepsList.append("{{p20Step}}")
pr.stepsList = stepsList

###########################################
# Configurable and fixed parameters
###########################################

pr.appendName = "{{WorkflowAppendName#GENERAL: Workflow string to append to production name#1}}"

w = "{{w#----->WORKFLOW: choose what to do#}}"
pr.prodsTypeList = "{{r#---List here which prods you want to create#DataSwimming,DataStripping,Merge}}".split(",")
p = "{{p#--now map steps to productions#}}"
p1 = "{{p1#-Production 1 steps (e.g. 1,2,3 = first, second, third)#1,2,3}}"
if p1:
    p1 = p1.split(",")
    p1 = [int(x) for x in p1]
    pr.stepsInProds.append(p1)
p2 = "{{p2#-Production 2 steps (e.g. 4,5)#4,5}}"
if p2:
    p2 = p2.split(",")
    p2 = [int(x) for x in p2]
    pr.stepsInProds.append(p2)
else:
    pr.stepsInProds.append([])
p3 = "{{p3#-Production 3 steps (e.g. 6)#}}"
if p3:
    p3 = p3.split(",")
    p3 = [int(x) for x in p3]
    pr.stepsInProds.append(p3)
else:
    pr.stepsInProds.append([])

localTestFlag = ast.literal_eval("{{localTestFlag#GENERAL: Set True for local test#False}}")
validationFlag = ast.literal_eval("{{validationFlag#GENERAL: Set True for validation prod#False}}")

# workflow params for all productions
pr.startRun = int("{{startRun#GENERAL: run start, to set the start run#0}}")
pr.endRun = int("{{endRun#GENERAL: run end, to set the end of the range#0}}")
pr.runsList = "{{runsList#GENERAL: discrete list of run numbers (do not mix with start/endrun)#}}"
pr.targets = ["{{WorkflowDestination#GENERAL: Workflow destination site e.g. LCG.CERN.cern#}}"] * len(pr.prodsTypeList)
extraOptions = "{{extraOptions#GENERAL: extra options as python dict stepNumber:options#}}"
if extraOptions:
    pr.extraOptions = ast.literal_eval(extraOptions)
eventTimeouts = "{{eventTimeouts#GENERAL: event timeouts as python dict stepNumber:seconds#}}"
if eventTimeouts:
    pr.eventTimeouts = ast.literal_eval(eventTimeouts)
pr.derivedProduction = int("{{AncestorProd#GENERAL: ancestor prod to be derived#0}}")
pr.previousProdID = int("{{previousProdID#GENERAL: previous prod ID (for BK query)#0}}")
modulesList = "{{modulesList#GENERAL: custom modules list#}}"
enablePopularityReport = ast.literal_eval("{{popularityReport#GENERAL: enable popularity report#False}}")
pr.visibility = "{{visibilityFlag#GENERAL: visibility flag for input BK Query#Yes}}"

# p1 params
p1Plugin = "{{p1PluginType#PROD-P1: production plugin name#LHCbStandard}}"
p1Priority = int("{{p1Priority#PROD-P1: priority#2}}")
p1CPU = "{{p1MaxCPUTime#PROD-P1: Max CPU time in secs#1000000}}"
p1GroupSize = "{{p1GroupSize#PROD-P1: Group size or number of files per job#1}}"
p1DataSE = "{{p1DataSE#PROD-P1: Output Data Storage Element#Tier1-DST}}"
try:
    p1DataSESpecial = ast.literal_eval('{{p1DataSESpecial#PROD-P1: Special SEs per file type, e.g. {"T1":"SE1"}#}}')
except SyntaxError:
    p1DataSESpecial = {}
p1Policy = "{{p1Policy#PROD-P1: data policy (download or protocol)#download}}"
p1RemoveInputs = ast.literal_eval("{{p1RemoveInputs#PROD-P1: removeInputs flag#False}}")
p1StepMask = "{{p1StepMask#PROD-P1: step output to save, semicolon separated (default is last)#}}"
p1FileMask = "{{p1FileMask#PROD-P1: file types in output to save, semicolon separated (default is all)#}}"
p1NoBKQuery = ast.literal_eval("{{P1NoBKQuery#PROD-P1: run without an input Bookkeeping Query#False}}")
p1multicoreFlag = "{{p1MulticoreFLag#PROD-P1: multicore flag#True}}"
p1NumberOfProcessors = "{{p1NumberOfProcessors#PROD-P1: jobs min/max n of processors#0,0}}"
p1outputMode = "{{p1OutputMode#PROD-P1: output mode#Local}}"
p1eventsRequested = "{{p1EventsRequested#PROD-P1: events requested (-1 = ALL)#-1}}"
p1ancestorDepth = int("{{p1AncestorDepth#PROD-P1: ancestor depth#0}}")
try:
    p1compressionLvl = "{{p1CompressionLevel#PROD-P1: Compression Level per step, e.g. LOW,HIGH]#LOW}}"
except SyntaxError:
    p1compressionLvl = []
p1OutputVisFlag = ast.literal_eval(
    '{{p1OutputVisFlag#PROD-P1: Visibility flag of output files (a dict {"n step":"flag"})#{"1":"N"} }}'
)
p1OutputVisFlag = fillVisList(p1OutputVisFlag, len(pr.stepsInProds[0]))
try:
    p1VisFlagSpecial = "{{p1OutputVisFlagSpecial"
    p1VisFlagSpecial += '#PROD-P1: Special Visibility flag of output files (a dict {"n step":{"FType":"flag" } })#}}'
    p1OutputVisFlagSpecial = ast.literal_eval(p1VisFlagSpecial)
except SyntaxError:
    p1OutputVisFlagSpecial = {}

# p2 params
p2Plugin = "{{p2PluginType#PROD-P2: production plugin name#LHCbStandard}}"
p2Priority = int("{{p2Priority#PROD-P2: priority#2}}")
p2CPU = "{{p2MaxCPUTime#PROD-P2: Max CPU time in secs#1000000}}"
p2GroupSize = "{{p2GroupSize#PROD-P2: Group Size#1}}"
p2DataSE = "{{p2DataSE#PROD-P2: Output Data Storage Element#Tier1-DST}}"
try:
    p2DataSESpecial = ast.literal_eval('{{p2DataSESpecial#PROD-P2: Special SEs per file type, e.g. {"T1":"SE1"}#}}')
except SyntaxError:
    p2DataSESpecial = {}
p2Policy = "{{p2Policy#PROD-P2: data policy (download or protocol)#download}}"
p2RemoveInputs = ast.literal_eval("{{p2RemoveInputs#PROD-P2: removeInputs flag#False}}")
p2StepMask = "{{p2StepMask#PROD-P2: step output to save, semicolon separated (default is last#}}"
p2FileMask = "{{p2FileMask#PROD-P2: file types in output to save, semicolon separated (default is all)#}}"
p2multicoreFlag = "{{p2MulticoreFLag#PROD-P2: multicore flag#True}}"
p2NumberOfProcessors = "{{p2NumberOfProcessors#PROD-P2: jobs min/max n of processors#0,0}}"
p2outputMode = "{{p2OutputMode#PROD-P2: output mode#Local}}"
p2eventsRequested = "{{p2EventsRequested#PROD-P2: events requested (-1 = ALL)#-1}}"
p2ancestorDepth = int("{{p2AncestorDepth#PROD-P2: ancestor depth#0}}")
try:
    p2compressionLvl = "{{p2CompressionLevel#PROD-P2: Compression Level per step, e.g. LOW,HIGH#LOW}}"
except SyntaxError:
    p2compressionLvl = []
p2OutputVisFlag = ast.literal_eval(
    '{{p2OutputVisFlag#PROD-P2: Visibility flag of output files (a dict {"n step":"flag"})#{"2":"Y"} }}'
)
p2OutputVisFlag = fillVisList(p2OutputVisFlag, len(pr.stepsInProds[1]))
try:
    p2VisFlagSpecial = "{{p2OutputVisFlagSpecial"
    p2VisFlagSpecial += '#PROD-P2: Special Visibility flag of output files (a dict {"n step":{"FType":"flag" } })#}}'
    p2OutputVisFlagSpecial = ast.literal_eval(p2VisFlagSpecial)
except SyntaxError:
    p2OutputVisFlagSpecial = {}

# p3 params
p3Plugin = "{{p3PluginType#PROD-P3: production plugin name#LHCbStandard}}"
p3Priority = int("{{p3Priority#PROD-P3: priority#2}}")
p3CPU = "{{p3MaxCPUTime#PROD-P3: Max CPU time in secs#1000000}}"
p3GroupSize = "{{p3GroupSize#PROD-P3: Group Size#1}}"
p3DataSE = "{{p3DataSE#PROD-P3: Output Data Storage Element#Tier1-DST}}"
try:
    p3DataSESpecial = ast.literal_eval('{{p3DataSESpecial#PROD-P3: Special SEs per file type, e.g. {"T1":"SE1"}#}}')
except SyntaxError:
    p3DataSESpecial = {}
p3Policy = "{{p3Policy#PROD-P3: data policy (download or protocol)#download}}"
p3RemoveInputs = ast.literal_eval("{{p3RemoveInputs#PROD-P3: removeInputs flag#False}}")
p3StepMask = "{{p3StepMask#PROD-P3: step output to save, semicolon separated (default is last#}}"
p3FileMask = "{{p3FileMask#PROD-P3: file types in output to save, semicolon separated (default is all)#}}"
p3multicoreFlag = "{{p3MulticoreFLag#PROD-P3: multicore flag#True}}"
p3NumberOfProcessors = "{{p3NumberOfProcessors#PROD-P3: jobs min/max n of processors#0,0}}"
p3outputMode = "{{p3OutputMode#PROD-P3: output mode#Any}}"
p3eventsRequested = "{{p3EventsRequested#PROD-P3: events requested (-1 = ALL)#-1}}"
p3ancestorDepth = int("{{p3AncestorDepth#PROD-P3: ancestor depth#0}}")
try:
    p3compressionLvl = '{{p3CompressionLevel#PROD-P3: Compression Level per step, e.g. ["LOW","HIGH"]#HIGH}}'
except SyntaxError:
    p3compressionLvl = []
p3OutputVisFlag = ast.literal_eval(
    '{{p3OutputVisFlag#PROD-P3: Visibility flag of output files (a dict {"n step":"flag"})#{"3":"Y"} }}'
)
p3OutputVisFlag = fillVisList(p3OutputVisFlag, len(pr.stepsInProds[2]))
try:
    p3VisFlagSpecial = "{{p3OutputVisFlagSpecial"
    p3VisFlagSpecial += '#PROD-P3: Special Visibility flag of output files (a dict {"n step":{"FType":"flag" } })#}}'
    p3OutputVisFlagSpecial = ast.literal_eval(p3VisFlagSpecial)
except SyntaxError:
    p3OutputVisFlagSpecial = {}

parentReq = "{{_parent}}"
if not parentReq:
    pr.requestID = "{{ID}}"
else:
    pr.requestID = parentReq

pr.prodGroup = "{{inProPass}}" + "/" + "{{pDsc}}"
# used in case of a test e.g. certification etc.
pr.configName = "{{configName}}"
pr.configVersion = "{{configVersion}}"
pr.outConfigName = pr.configName
# Other parameters from the request page
pr.dqFlag = "{{inDataQualityFlag}}"  # UNCHECKED
pr.dataTakingConditions = "{{simDesc}}"
pr.processingPass = "{{inProPass}}"
if p1[0] == 1 and pr.prodsTypeList[0].lower() not in ("mcsimulation", "mcfastsimulation"):
    pr.bkFileType = "{{inFileType}}"
    pr.bkQueries = ["Full"] if not p1NoBKQuery else [""]
elif p1[0] == 1 and pr.prodsTypeList[0].lower() in ("mcsimulation", "mcfastsimulation"):
    pr.bkQueries = [""]
    pr.prodGroup = "{{pDsc}}"
elif p1[0] != 1 and pr.prodsTypeList[0].lower() not in ("mcsimulation", "mcfastsimulation"):
    pr.bkQueries = ["fromPreviousProd"]
    if not pr.previousProdID:
        gLogger.error("Please specify an input production")
        DIRACexit(2)
pr.eventType = "{{eventType}}"

if modulesList:
    pr.modulesList = modulesList.replace(" ", "").split(",")

if localTestFlag:
    pr.testFlag = True
    pr.publishFlag = False
    pr.prodsToLaunch = [1]

if validationFlag:
    pr.outConfigName = "validation"

if enablePopularityReport:
    pr.extraModulesList = ["FileUsage"]

inputDataList = []
if not pr.publishFlag:
    testData = "LFN:/lhcb/LHCb/Collision11/CHARMCOMPLETEEVENT.DST"
    testData += "/00012586/0000/00012586_00000706_1.charmcompleteevent.dst"
    inputDataList.append(testData)
    p1Policy = "protocol"
    evtsPerJob = "5"
    bkScriptFlag = True
else:
    bkScriptFlag = False

# In case we want just to test, we publish in the certification/test part of the BKK
if currentSetup == "LHCb-Certification" or pr.testFlag:
    pr.outConfigName = "certification"
    pr.configVersion = "test"

if pr.testFlag:
    pr.startRun = 75336
    pr.endRun = 75340
    p1CPU = "100000"
    pr.events = ["2000"]
    pr.dataTakingConditions = "Beam3500GeV-VeloClosed-MagDown"
    pr.processingPass = "Real Data/Reco12/Stripping17"
    pr.bkFileType = "CHARMCOMPLETEEVENT.DST"
    pr.eventType = "90000000"
    pr.dqFlag = "ALL"

if not p1StepMask:
    p1StepMask = len(pr.stepsInProds[0])
if not p2StepMask:
    try:
        p2StepMask = len(pr.stepsInProds[1])
    except IndexError:
        p2StepMask = ""
if not p3StepMask:
    try:
        p3StepMask = len(pr.stepsInProds[2])
    except IndexError:
        p3StepMask = ""

p1NumberOfProcessors = tuple(int(x) for x in p1NumberOfProcessors.replace(" ", "").split(","))
p2NumberOfProcessors = tuple(int(x) for x in p2NumberOfProcessors.replace(" ", "").split(","))
p3NumberOfProcessors = tuple(int(x) for x in p3NumberOfProcessors.replace(" ", "").split(","))

pr.outputSEs = [x for x in [p1DataSE, p2DataSE, p3DataSE] if x != ""]
pr.specialOutputSEs = [p1DataSESpecial, p2DataSESpecial, p3DataSESpecial]
pr.removeInputsFlags = [p1RemoveInputs, p2RemoveInputs, p3RemoveInputs][0 : len(pr.prodsTypeList)]
pr.priorities = [p1Priority, p2Priority, p3Priority][0 : len(pr.prodsTypeList)]
pr.cpus = [p1CPU, p2CPU, p3CPU][0 : len(pr.prodsTypeList)]
pr.groupSizes = [p1GroupSize, p2GroupSize, p3GroupSize][0 : len(pr.prodsTypeList)]
pr.plugins = [p1Plugin, p2Plugin, p3Plugin][0 : len(pr.prodsTypeList)]
pr.inputs = [inputDataList, [], []][0 : len(pr.prodsTypeList)]
pr.inputDataPolicies = [p1Policy, p2Policy, p3Policy][0 : len(pr.prodsTypeList)]
pr.outputFileSteps = [p1StepMask, p2StepMask, p3StepMask]
pr.outputFileMasks = [p1FileMask, p2FileMask, p3FileMask]
pr.multicore = [p1multicoreFlag, p2multicoreFlag, p3multicoreFlag]
pr.processors = [p1NumberOfProcessors, p2NumberOfProcessors, p3NumberOfProcessors]
pr.outputModes = [p1outputMode, p2outputMode, p3outputMode]
pr.events = [p1eventsRequested, p2eventsRequested, p3eventsRequested]
pr.ancestorDepths = [p1ancestorDepth, p2ancestorDepth, p3ancestorDepth]
# pr.compressionLvl = [p1compressionLvl] * len( pr.stepsInProds[0] ) +\
#                    [p2compressionLvl] * len( pr.stepsInProds[1] ) +\
#                    [p3compressionLvl] * len( pr.stepsInProds[2] )
pr.compressionLvl = (
    [lvl for lvl in p1compressionLvl.split(",")]
    + [lvl for lvl in p2compressionLvl.split(",")]
    + [lvl for lvl in p3compressionLvl.split(",")]
)
pr.outputVisFlag = [p1OutputVisFlag, p2OutputVisFlag, p3OutputVisFlag]
pr.specialOutputVisFlag = [p1OutputVisFlagSpecial, p2OutputVisFlagSpecial, p3OutputVisFlagSpecial]

pr.buildAndLaunchRequest()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unittest for: LHCbDIRAC.AccountingSystem.private.Plotters.PopularityPlotter.

PopularityPlotter.__bases__:
  DIRAC.AccountingSystem.private.Plotters.BaseReporter

We are assuming there is a solid test of __bases__, we are not testing them
here and assuming they work fine.

IMPORTANT: the test MUST be pylint compliant !
"""
# pylint: disable=protected-access

import unittest
from decimal import Decimal

from unittest import mock


class PopularityPlotterTestCase(unittest.TestCase):
    """PopularityPlotterTestCase."""

    moduleTested = None
    classsTested = None

    def mockModuleTested(self, moduleTested):
        """Used to not redo the mocking done on the parent class ( if any )"""

        # Tries to get the mocks of the parent TestCases ( if any )
        for baseClass in self.__class__.__bases__:
            try:
                # pylint: disable=no-member
                moduleTested = baseClass.mockModuleTested(moduleTested)
            except TypeError:
                continue

        # And then makes its own mock
        class MockPopularity:
            # pylint: disable=missing-docstring,no-init
            definitionKeyFields = (
                "DataType",
                "Activity",
                "FileType",
                "Production",
                "ProcessingPass",
                "Conditions",
                "EventType",
                "StorageElement",
            )

        moduleTested.Popularity = mock.Mock(return_value=MockPopularity())

        return moduleTested

    def setUp(self):
        """Setup the test case."""

        import LHCbDIRAC.AccountingSystem.private.Plotters.PopularityPlotter as moduleTested

        self.moduleTested = self.mockModuleTested(moduleTested)
        self.classsTested = self.moduleTested.PopularityPlotter

    def tearDown(self):
        """Tear down the test case."""

        del self.moduleTested
        del self.classsTested


# ...............................................................................


class PopularityPlotterUnitTest(PopularityPlotterTestCase):
    """PopularityPlotterUnitTest.

    <constructor>
     - test_instantiate
    <class variables>
     - test_typeName
     - test_typeKeyFields
     - test_reportDataUsageName
     - test_reportNormalizedDataUsageName
    <methods>
     - test_reportDataUsage
     - test_reportNormalizedDataUsage
     - test_plotDataUsage
     - test_plotNormalizedDataUsage
    """

    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""
        obj = self.classsTested(None, None)
        self.assertEqual("PopularityPlotter", obj.__class__.__name__)

    def test_typeName(self):
        """test the class variable "_typeName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._typeName, "Popularity")

    def test_typeKeyFields(self):
        """test the class variable "_typeKeyFields"."""
        obj = self.classsTested(None, None)
        self.assertEqual(
            obj._typeKeyFields,
            [
                "DataType",
                "Activity",
                "FileType",
                "Production",
                "ProcessingPass",
                "Conditions",
                "EventType",
                "StorageElement",
            ],
        )

    def test_reportDataUsageName(self):
        """test the class variable "_reportDataUsageName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportDataUsageName, "Data Usage")

    def test_reportNormalizedDataUsageName(self):
        """test the class variable "_reportNormalizedDataUsageName"."""
        obj = self.classsTested(None, None)
        self.assertEqual(obj._reportNormalizedDataUsageName, "Normalized Data Usage")

    def test_reportDataUsage(self):
        """test the method "_reportDataUsage"."""

        mockAccountingDB = mock.MagicMock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        res = obj._reportDataUsage(
            {"groupingFields": [0, ["mehh"], "blah"], "startTime": "startTime", "endTime": "endTime", "condDict": {}}
        )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportDataUsage(
            {"groupingFields": [0, ["mehh"], "blah"], "startTime": "startTime", "endTime": "endTime", "condDict": {}}
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"], {"graphDataDict": {}, "data": {}, "unit": "files", "granularity": "BucketLength"}
        )

        mockedData = (
            ("90000000", 1355616000, 86400, Decimal("123456.789")),
            ("90000000", 1355702400, 86400, Decimal("78901.2345")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportDataUsage(
            {
                "groupingFields": ("%s", ["EventType"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"EventType": "90000000"},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {"90000000": {1355616000: 123.456789, 1355702400: 78.901234500000001}},
                "data": {"90000000": {1355616000: 123456.789, 1355702400: 78901.234500000006}},
                "unit": "kfiles",
                "granularity": 86400,
            },
        )

    def test_reportNormalizedDataUsage(self):
        """test the method "_reportNormalizedDataUsage"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        #     res = obj._reportNormalizedDataUsage( { 'groupingFields' : [ 0, [ 'mehh' ], 'blah' ],
        #                                             'startTime'      : 'startTime',
        #                                             'endTime'        : 'endTime',
        #                                             'condDict'       : {}
        #                                            } )
        #     self.assertEqual( res[ 'OK' ], False )
        #     self.assertEqual( res[ 'Message' ], 'No connection' )

        # Changed mocked to run over different lines of code
        mockAccountingDB._getConnection.return_value = {"OK": True, "Value": []}
        res = obj._reportNormalizedDataUsage(
            {"groupingFields": [0, ["mehh"], "blah"], "startTime": "startTime", "endTime": "endTime", "condDict": {}}
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"], {"graphDataDict": {}, "data": {}, "unit": "files", "granularity": "BucketLength"}
        )

        mockedData = (
            ("90000000", 1355616000, 86400, Decimal("123456.789")),
            ("90000000", 1355702400, 86400, Decimal("78901.2345")),
            ("90000001", 1355616000, 86400, Decimal("223456.789")),
            ("90000001", 1355702400, 86400, Decimal("148901.2345")),
        )
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": mockedData}
        mockAccountingDB.calculateBucketLengthForTime.return_value = 86400

        res = obj._reportNormalizedDataUsage(
            {
                "groupingFields": ("%s", ["EventType"]),
                "startTime": 1355663249.0,
                "endTime": 1355749690.0,
                "condDict": {"StorageElement": "CERN"},
            }
        )
        self.assertEqual(res["OK"], True)
        self.assertEqual(
            res["Value"],
            {
                "graphDataDict": {
                    "90000001": {1355616000: 223.45678899999999, 1355702400: 148.90123449999999},
                    "90000000": {1355616000: 123.456789, 1355702400: 78.901234500000001},
                },
                "data": {
                    "90000001": {1355616000: 223456.78899999999, 1355702400: 148901.23449999999},
                    "90000000": {1355616000: 123456.789, 1355702400: 78901.234500000006},
                },
                "unit": "kfiles",
                "granularity": 86400,
            },
        )


# ...............................................................................


class PopularityPlotterUnitTestCrashes(PopularityPlotterTestCase):
    """PopularityPlotterUnitTestCrashes.

    <constructor>
     - test_instantiate
    <class variables>
    <methods>
    - test_reportDataUsage
    - test_reportNormalizedDataUsage
    - test_plotDataUsage
    - test_plotNormalizedDataUsage
    """

    def test_instantiate(self):
        """test the constructor."""

        self.assertRaises(TypeError, self.classsTested)
        self.assertRaises(TypeError, self.classsTested, None)
        self.assertRaises(TypeError, self.classsTested, None, None, None, None)

        self.assertRaises(TypeError, self.classsTested, extraArgs=None)
        self.assertRaises(TypeError, self.classsTested, None, extraArgs=None)
        self.assertRaises(TypeError, self.classsTested, None, None, None, extraArgs=None)

    def test_reportDataUsage(self):
        """test the method "_reportDataUsage"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportDataUsage, {})
        self.assertRaises(IndexError, obj._reportDataUsage, {"groupingFields": []})
        self.assertRaises(TypeError, obj._reportDataUsage, {"groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportDataUsage, {"groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportDataUsage, {"groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportDataUsage, {"groupingFields": ["1", [2]]})
        self.assertRaises(KeyError, obj._reportDataUsage, {"groupingFields": ["1", [2, 2]], "startTime": None})
        self.assertRaises(
            KeyError, obj._reportDataUsage, {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None}
        )
        self.assertRaises(
            TypeError,
            obj._reportDataUsage,
            {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_reportNormalizedDataUsage(self):
        """test the method "_reportNormalizedDataUsage"."""

        mockAccountingDB = mock.Mock()
        mockAccountingDB._getConnection.return_value = {"OK": False, "Message": "No connection"}
        mockAccountingDB.retrieveBucketedData.return_value = {"OK": True, "Value": []}
        mockAccountingDB.calculateBucketLengthForTime.return_value = "BucketLength"
        obj = self.classsTested(mockAccountingDB, None)

        self.assertRaises(KeyError, obj._reportNormalizedDataUsage, {})
        self.assertRaises(IndexError, obj._reportNormalizedDataUsage, {"groupingFields": []})
        self.assertRaises(TypeError, obj._reportNormalizedDataUsage, {"groupingFields": [1, 2]})
        self.assertRaises(TypeError, obj._reportNormalizedDataUsage, {"groupingFields": [1, [2]]})
        self.assertRaises(TypeError, obj._reportNormalizedDataUsage, {"groupingFields": ["1", "2"]})
        self.assertRaises(KeyError, obj._reportNormalizedDataUsage, {"groupingFields": ["1", [2]]})
        self.assertRaises(
            KeyError, obj._reportNormalizedDataUsage, {"groupingFields": ["1", [2, 2]], "startTime": None}
        )
        self.assertRaises(
            KeyError,
            obj._reportNormalizedDataUsage,
            {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None},
        )
        self.assertRaises(
            TypeError,
            obj._reportNormalizedDataUsage,
            {"groupingFields": ["1", [2, 2]], "startTime": None, "endTime": None, "condDict": None},
        )

    def test_plotDataUsage(self):
        """test the method "_plotDataUsage"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotDataUsage, None, None, None)
        self.assertRaises(KeyError, obj._plotDataUsage, {}, None, None)
        self.assertRaises(KeyError, obj._plotDataUsage, {"startTime": "startTime"}, None, None)
        self.assertRaises(TypeError, obj._plotDataUsage, {"startTime": "startTime", "endTime": "endTime"}, None, None)
        self.assertRaises(KeyError, obj._plotDataUsage, {"startTime": "startTime", "endTime": "endTime"}, {}, None)
        self.assertRaises(
            KeyError,
            obj._plotDataUsage,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotDataUsage,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotDataUsage,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )

    def test_plotNormalizedDataUsage(self):
        """test the method "_plotNormalizedDataUsage"."""

        obj = self.classsTested(None, None)
        self.assertRaises(TypeError, obj._plotNormalizedDataUsage, None, None, None)
        self.assertRaises(KeyError, obj._plotNormalizedDataUsage, {}, None, None)
        self.assertRaises(KeyError, obj._plotNormalizedDataUsage, {"startTime": "startTime"}, None, None)
        self.assertRaises(
            TypeError, obj._plotNormalizedDataUsage, {"startTime": "startTime", "endTime": "endTime"}, None, None
        )
        self.assertRaises(
            KeyError, obj._plotNormalizedDataUsage, {"startTime": "startTime", "endTime": "endTime"}, {}, None
        )
        self.assertRaises(
            KeyError,
            obj._plotNormalizedDataUsage,
            {"startTime": "startTime", "endTime": "endTime"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotNormalizedDataUsage,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity"},
            None,
        )
        self.assertRaises(
            KeyError,
            obj._plotNormalizedDataUsage,
            {"startTime": "startTime", "endTime": "endTime", "grouping": "grouping"},
            {"granularity": "granularity", "graphDataDict": "graphDataDict"},
            None,
        )


################################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(PopularityPlotterTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PopularityPlotterUnitTest))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PopularityPlotterUnitTestCrashes))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Given a (list of) LFNs and SEs, check for existence of a file and register
in the FC if the file exists."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    dmScript.registerSiteSwitches()

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ...  [LFN1[,LFN2,[...]]] [--SE] Dest[,Dest2[,...]] ",
                "Arguments:",
                "  Dest:     Valid DIRAC SE(s)",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeRegisterBK2FC
    from DIRAC import exit

    exit(executeRegisterBK2FC(dmScript))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DataProcessingProgressAgent."""
import datetime
import os
import time
import shutil

from DIRAC import S_OK, gConfig, gLogger
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Base.AgentModule import AgentModule

from LHCbDIRAC.ProductionManagementSystem.Client.ProcessingProgress import ProcessingProgress, HTMLProgressTable
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery

AGENT_NAME = "ProductionManagement/DataProcessingProgressAgent"


class DataProcessingProgressAgent(AgentModule):
    #############################################################################
    def initialize(self):
        """Sets default values."""

        self.pollingTime = self.am_getOption("PollingTime", 6 * 60 * 60)
        self.printResult = self.am_getOption("Verbose", False)
        if self.printResult:
            gLogger.setLevel("VERBOSE")
        self.workDirectory = self.am_getWorkDirectory()
        self.statCollector = ProcessingProgress(os.path.join(self.workDirectory, "dirac-production-stats.pkl"))
        self.uploadDirectory = self.am_getOption("UploadDirectory", None)

        # Get back the loop number
        self.cacheFile = os.path.join(self.workDirectory, "cacheFile")
        try:
            cfile = open(self.cacheFile)
            self.iterationNumber = int(cfile.read())
            cfile.close()
        except BaseException:
            self.iterationNumber = 0

        # Get the list of processing passes
        reportList = self.am_getSection("ProgressReports")
        self.progressReports = {}
        for reportName in reportList:
            optionPath = os.path.join("ProgressReports", reportName)
            processingPasses = self.am_getOption(os.path.join(optionPath, "ProcessingPass"), [])
            conditions = self.am_getOption(os.path.join(optionPath, "ConditionDescription"), [])
            bkConfig = self.am_getOption(os.path.join(optionPath, "BKConfig"), "/LHCb/Collision11")
            eventType = self.am_getOption(os.path.join(optionPath, "EventType"), 90000000)
            fileType = self.am_getOption(os.path.join(optionPath, "FileType"), "BHADRON.DST")
            report = {"ConditionDescription": conditions}
            report["Frequency"] = self.am_getOption(os.path.join(optionPath, "Frequency"), 1)
            report["HTMLFile"] = self.am_getOption(
                os.path.join(optionPath, "HTMLFile"), reportName.replace(".", "") + "-Progress" + os.path.extsep + "htm"
            )
            report["BKQuery"] = []
            report["ClearCache"] = self.am_getOption(os.path.join(optionPath, "ClearCache"), [])
            for processingPass in processingPasses:
                bkPath = os.path.join(bkConfig, "*/Real Data", processingPass, str(eventType), fileType)
                bkQuery = BKQuery(bkPath, visible=False)
                if not bkQuery:
                    self.log.error(f"Cannot build bkQuery for {bkPath}")
                else:
                    report["BKQuery"].append(bkQuery)
            self.progressReports[reportName.replace(".", "/")] = report

        self.log.info("List of progress reports:")
        self.previousProdStats = {}
        for reportName in self.progressReports:
            printStr = f"{reportName}: "
            for key, value in self.progressReports[reportName].items():
                if key != "BKQuery":
                    printStr += f", {key} : {str(value)}"
                else:
                    printStr += f", {key} :"
                    for bkQuery in self.progressReports[reportName]["BKQuery"]:
                        printStr += f" - {bkQuery.getPath()}"
            self.log.info(printStr)
        return S_OK()

    def execute(self):
        self.log.info("Now getting progress of processing (iteration %d)..." % self.iterationNumber)

        for reportName in sorted(self.progressReports):
            htmlTable = HTMLProgressTable(reportName.replace(".", "/"))
            reportLen = len(reportName) + 4
            self.log.info(f"\n{reportLen * '*'}\n* {reportName} *\n{reportLen * '*'}")
            report = self.progressReports[reportName]
            # Skip all by each "frequency" loop
            if report["Frequency"] == 0 or (self.iterationNumber % report["Frequency"]) != 0:
                self.log.info(f"Skipping this iteration for {reportName}")
                continue
            self.statCollector.setClearCache(report["ClearCache"])
            summaryProdStats = []
            printOutput = ""
            outputHTML = os.path.join(self.workDirectory, report["HTMLFile"])
            for cond in report["ConditionDescription"]:
                prodStats = 4 * [None]
                for bkQuery in report["BKQuery"]:
                    bkPath = bkQuery.getPath()
                    if not bkQuery.getQueryDict():
                        continue
                    self.log.info(
                        "\n=========================\nBookkeeping query %s\n========================="
                        % bkPath.replace("*", cond)
                    )
                    bkQuery.setConditions(cond)
                    stats = self.statCollector.getFullStats(bkQuery, printResult=self.printResult)
                    processingPass = bkQuery.getProcessingPass().split("/")
                    for ind in range(len(prodStats)):
                        if not prodStats[ind]:
                            prodStats[ind] = stats[ind]
                        else:
                            prodStats[ind] += stats[ind]
                summaryProdStats.append(prodStats)
                if self.printResult:
                    printOutput += self.statCollector.outputResults(cond, reportName.replace(".", "/"), prodStats)
                htmlTable.writeHTML(cond, prodStats)

            if self.printResult:
                lines = printOutput.split("\n")
                for line in lines:
                    self.log.info(line)

            if len(summaryProdStats) > 1:
                htmlTable.writeHTMLSummary(summaryProdStats)
            if reportName not in self.previousProdStats:
                x = self.statCollector.getPreviousStats(reportName)
                if x:
                    self.previousProdStats[reportName] = x
            if reportName in self.previousProdStats:
                htmlTable.writeHTMLDifference(summaryProdStats, self.previousProdStats[reportName])
            else:
                print(reportName, "not in previous stats")
            self.previousProdStats[reportName] = {"Time": time.ctime(time.time()), "ProdStats": summaryProdStats}
            self.statCollector.setPreviousStats(reportName, self.previousProdStats[reportName])
            try:
                fOpen = open(outputHTML, "w")
                fOpen.write(f"<head>\n<title>Progress of {bkQuery.getProcessingPass()}</title>\n</title>\n")
                fOpen.write(str(htmlTable.getTable()))
                fOpen.close()
                print("Successfully wrote HTML file", outputHTML)
                self.uploadHTML(outputHTML)
            except BaseException:
                print("Failed to write HTML file", outputHTML)

        # Save the loop number
        self.iterationNumber += 1
        fOpen = open(self.cacheFile, "w")
        fOpen.write(str(self.iterationNumber))
        fOpen.close()
        return S_OK()

    def uploadHTML(self, htmlFile):
        if not self.uploadDirectory:
            return
        try:
            uploadDirBase = os.path.join(self.uploadDirectory, "Daily", str(datetime.datetime.today()).split()[0])
            mkDir(uploadDirBase)
            i = 0
            while True:
                uploadDir = os.path.join(uploadDirBase, str(i))
                mkDir(uploadDir)
                uploadedFile = os.path.join(uploadDir, os.path.basename(htmlFile))
                if not os.path.exists(uploadedFile):
                    break
                i += 1
            shutil.copy(htmlFile, uploadedFile)
            remoteLink = os.path.join(self.uploadDirectory, os.path.basename(htmlFile))
            if os.path.exists(remoteLink):
                os.remove(remoteLink)
            os.symlink(uploadedFile, remoteLink)
            print(htmlFile, "copied to", uploadedFile, "and link set at", remoteLink)
        except BaseException:
            print("Failed to upload", htmlFile, "to", self.uploadDirectory)

    def am_getSection(self, section):
        res = gConfig.getSections(f"{self.am_getModuleParam('section')}/{section}")
        if res["OK"]:
            return res["Value"]
        else:
            return []


################################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for plugins."""
# imports
import unittest
import importlib
from unittest.mock import MagicMock

from DIRAC import gLogger

# sut
from LHCbDIRAC.TransformationSystem.Agent.TransformationPlugin import TransformationPlugin


paramsBase = {
    "AgentType": "Automatic",
    "DerivedProduction": "0",
    "FileMask": "",
    "GroupSize": 1,
    "InheritedFrom": 0,
    "JobType": "MCSimulation",
    "MaxNumberOfTasks": 0,
    "OutputDirectories": "['/lhcb/MC/20', '/lhcb/debug/20']",
    "OutputLFNs": "{'LogTargetPath': ['/lhcb/9.tar'], 'LogFilePath': ['/lhcb/9']}",
    "Priority": "0",
    "SizeGroup": "1",
    "Status": "Active",
    "TransformationID": 1080,
    "Type": "MCSimulation",
    "outputDataFileMask": "GAUSSHIST;ALLSTREAMS.DST",
}

data = {
    "/this/is/at.1": ["SE1"],
    "/this/is/at.2": ["SE2"],
    "/this/is/also/at.2": ["SE2"],
    "/this/is/at.12": ["SE1", "SE2"],
    "/this/is/also/at.12": ["SE1", "SE2"],
    "/this/is/at_123": ["SE1", "SE2", "SE3"],
    "/this/is/at_23": ["SE2", "SE3"],
    "/this/is/at_4": ["SE4"],
}


class PluginsTestCase(unittest.TestCase):
    """Base class for the Agents test cases."""

    def setUp(self):
        self.mockTC = MagicMock()
        self.mockDM = MagicMock()
        self.mockRM = MagicMock()
        self.mockBKK = MagicMock()
        self.mockCatalog = MagicMock()
        self.mockCatalog.getFileSize.return_value()
        self.tPlugin = importlib.import_module("LHCbDIRAC.TransformationSystem.Agent.TransformationPlugin")
        self.tPlugin.TransformationClient = self.mockTC
        self.tPlugin.DataManager = self.mockDM
        self.tPlugin.FileCatalog = self.mockCatalog
        self.tPlugin.rmClient = self.mockRM
        self.maxDiff = None
        se = importlib.import_module("DIRAC.TransformationSystem.Client.Utilities")
        mockSE = MagicMock()
        se.StorageElement = mockSE

        gLogger.setLevel("DEBUG")

    def tearDown(self):
        #     sys.modules.pop( 'DIRAC.Core.Base.AgentModule' )
        #     sys.modules.pop( 'DIRAC.TransformationSystem.Agent.TransformationAgent' )
        pass


class PluginsBaseSuccess(PluginsTestCase):
    def test__LHCbStandard(self):
        # no input data, active
        pluginStandard = TransformationPlugin(
            "LHCbStandard",
            transClient=self.mockTC,
            dataManager=self.mockDM,
            rmClient=self.mockRM,
            bkClient=self.mockBKK,
            fc=self.mockCatalog,
        )
        pluginStandard.setParameters(paramsBase)
        res = pluginStandard.run()
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], [])

        # input data, active
        pluginStandard = TransformationPlugin(
            "LHCbStandard",
            transClient=self.mockTC,
            dataManager=self.mockDM,
            rmClient=self.mockRM,
            bkClient=self.mockBKK,
            fc=self.mockCatalog,
        )
        pluginStandard.setParameters(paramsBase)
        pluginStandard.setInputData(data)
        params = {}
        params["TransformationID"] = 123
        params["Status"] = "Active"
        params["GroupSize"] = 2
        pluginStandard.setParameters(params)
        res = pluginStandard.run()
        self.assertTrue(res["OK"])
        self.assertTrue(len(res["Value"]) == 3)
        for t in res["Value"]:
            self.assertTrue(len(t[1]) <= 2)

        # input data, flush
        pluginStandard = TransformationPlugin(
            "LHCbStandard",
            transClient=self.mockTC,
            dataManager=self.mockDM,
            rmClient=self.mockRM,
            bkClient=self.mockBKK,
            fc=self.mockCatalog,
        )
        pluginStandard.setParameters(paramsBase)
        pluginStandard.setInputData(data)
        params = {}
        params["TransformationID"] = 123
        params["Status"] = "Flush"
        params["GroupSize"] = 2
        pluginStandard.setParameters(params)
        res = pluginStandard.run()
        self.assertTrue(res["OK"])
        self.assertTrue(len(res["Value"]) == 6)
        for t in res["Value"]:
            self.assertTrue(len(t[1]) <= 2)


#############################################################################
# Test Suite run
#############################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(PluginsTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PluginsBaseSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It interprets the XML reports and make a job, file, or replica object."""
from xml.parsers.expat import ExpatError
from xml.dom.minidom import parse, parseString
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.ReturnValues import convertToReturnValue, returnValueOrRaise
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from LHCbDIRAC.BookkeepingSystem.DB.OracleBookkeepingDB import OracleBookkeepingDB
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.FileParam import FileParam
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.Job.JobParameters import JobParameters
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.JobReader import JobReader
from LHCbDIRAC.BookkeepingSystem.Service.XMLReader.ReplicaReader import ReplicaReader
from LHCbDIRAC.BookkeepingSystem.DB.DataTakingConditionInterpreter import generateConditionDescription


class XMLFilesReaderManager:
    """XMLFilesReaderManager class."""

    #############################################################################

    def __init__(self):
        """initialize the member of class."""
        self.db = OracleBookkeepingDB()
        self.log = gLogger.getSubLogger("XMLFilesReaderManager")

    #############################################################################
    @staticmethod
    def readFile(filename):
        """reads an file content which format is XML."""
        try:
            with open(filename) as stream:
                doc = parse(stream)

            docType = doc.doctype  # job or replica
            xmltype = docType.name
        except NameError as ex:
            gLogger.error("XML reading error", filename)
            return S_ERROR(ex)

        return xmltype, doc, filename

    #############################################################################
    @staticmethod
    def smogInjectionSMOG2Gas(smogValue):
        """converts RunDB/CondDB SMOG value format into BK format (currently SMOG2 only or NoGas)."""
        fields = smogValue.split("_")
        type = fields[0]
        gas = fields[1] if len(fields) > 1 else ""
        unstable = fields[2] if len(fields) > 2 else ""
        if type == "NONE" or gas == "NONE":  # we set NoGas even when it is speicified for SMOG(1)
            return "NoGas"
        if type != "SMOG2":
            return "Unknown"  # register the value, means "we know it is unknown"
        if unstable not in ["", "Unstable"] or not gas:
            gLogger.error(f"SMOG value '{smogValue}' has unexpected format")
            return ""
        return gas.capitalize() + unstable

    #############################################################################
    def readXMLfromString(self, xmlString):
        """read the xml string."""
        try:
            doc = parseString(xmlString)

            docType = doc.doctype  # job or replica
            xmltype = docType.name

            if xmltype == "Replicas":
                replica = ReplicaReader().readReplica(doc, "IN Memory")
                result = self.processReplicas(replica)
                del replica
                return result
            elif xmltype == "Job":
                job = JobReader().readJob(doc, "IN Memory")
                result = self.processJob(job)
                del job
                return result
            else:
                self.log.error("unknown XML file!!!")
        except ExpatError as ex:
            self.log.error("XML reading error", repr(ex))
            self.log.exception()
            return S_ERROR(ex)

    #############################################################################
    def processJob(self, job):
        """interprets the xml content."""
        self.log.debug("Start Job Processing")

        # prepare for the insert, check the existence of the input files and retreive the fileid
        inputFiles = [inputFile.name for inputFile in job.inputFiles]
        if inputFiles:
            result = self.db.bulkgetIDsFromFilesTable(inputFiles)
            if not result["OK"]:
                return result
            if result["Value"]["Failed"]:
                self.log.error("The following files are not in the bkk", f"{','.join(result['Value']['Failed'])}")
                return S_ERROR("Files not in bkk")

            for inputFile in job.inputFiles:
                inputFile.fileID = int(result["Value"]["Successful"][inputFile.name]["FileId"])

        fileTypeCache = {}
        for outputfile in job.outputFiles:
            typeName = outputfile.type
            typeVersion = outputfile.version
            cachedTypeNameVersion = typeName + "<<" + typeVersion
            if cachedTypeNameVersion in fileTypeCache:
                self.log.debug(cachedTypeNameVersion + " in the cache!")
                typeID = fileTypeCache[cachedTypeNameVersion]
                outputfile.typeID = typeID
            else:
                result = self.db.checkFileTypeAndVersion(typeName, typeVersion)
                if not result["OK"]:
                    self.log.error("The [type:version] is missing", f"[{typeName}: {typeVersion}]")
                    return S_ERROR("[type:version] missing")

                self.log.debug(cachedTypeNameVersion + " added to the cache!")
                typeID = int(result["Value"])
                outputfile.typeID = typeID
                fileTypeCache[cachedTypeNameVersion] = typeID

            if (
                job.getParam("JobType") and job.getParam("JobType").value == "DQHISTOMERGING"
            ):  # all the merged histogram files have to be visible
                newFileParams = FileParam()
                newFileParams.name = "VisibilityFlag"
                newFileParams.value = "Y"
                outputfile.addFileParam(newFileParams)
                self.log.debug("The Merged histograms visibility flag has to be Y!")

            evtExists = False

            for param in outputfile.params:
                self.log.debug("ParamName check of " + str(param.name))

                if param.name == "EventType" and param.value:
                    result = self.db.checkEventType(int(param.value))
                    if not result["OK"]:
                        return S_ERROR(f"The event type {str(param.value)} is missing!")

                if param.name == "EventTypeId" and param.value:
                    result = self.db.checkEventType(int(param.value))
                    if not result["OK"]:
                        return S_ERROR(f"The event type {str(param.value)} is missing!")
                    evtExists = True

            if not evtExists and outputfile.type != "LOG":
                inputFiles = job.inputFiles

                if inputFiles:
                    fileName = inputFiles[0].name
                    res = self.db.getFileMetadata([fileName])
                    if not res["OK"]:
                        return res
                    fileMetadata = res["Value"]["Successful"].get(fileName)
                    if fileMetadata:
                        if "EventTypeId" in fileMetadata:
                            if outputfile.exists("EventTypeId"):
                                param = outputfile.getParam("EventTypeId")
                                param.value = str(fileMetadata["EventTypeId"])
                            else:
                                newFileParams = FileParam()
                                newFileParams.name = "EventTypeId"
                                newFileParams.value = str(fileMetadata["EventTypeId"])
                                outputfile.addFileParam(newFileParams)
                    else:
                        errMsg = f"Can not get the metadata of {fileName} file"
                        self.log.error(errMsg)
                        return S_ERROR(errMsg)

                elif job.getOutputFileParam("EventTypeId") is not None:
                    param = job.getOutputFileParam("EventTypeId")
                    newFileParams = FileParam()
                    newFileParams.name = "EventTypeId"
                    newFileParams.value = param.value
                    outputfile.addFileParam(newFileParams)

                else:
                    return S_ERROR("It can not fill the EventTypeId because there is no input files!")

        dqvalue = None
        infiles = job.inputFiles
        if not job.exists("RunNumber") and infiles:  # case of, e.g., MCReconstruction
            # Discover the run(s) and TCK(s) of the input files
            # One goal is to discover which dataquality to use

            res = self._getRunNumbersAndTCKs([jif.name for jif in job.inputFiles])
            if not res["OK"]:  # pylint: disable=invalid-sequence-index
                return res
            runNumbers, tcks = res["Value"]  # pylint: disable=invalid-sequence-index

            if len(runNumbers) > 1:
                self.log.debug("More than 1 run", ",".join(str(r) for r in runNumbers))
            elif len(runNumbers) == 1:
                runNumber = runNumbers.pop()
                self.log.debug("The output files of the job might inherit run", runNumber)
                newJobParams = JobParameters()
                newJobParams.name = "RunNumber"
                newJobParams.value = str(runNumber)
                job.addJobParams(newJobParams)

                if job.getParam("JobType") and job.getParam("JobType").value == "DQHISTOMERGING":
                    self.log.debug("DQ merging!")
                    retVal = self.db.getJobInfo(job.inputFiles[0].name)
                    if not retVal["OK"]:
                        return retVal
                    prod = retVal["Value"][0][18]
                    newJobParams = JobParameters()
                    newJobParams.name = "Production"
                    newJobParams.value = str(prod)
                    job.addJobParams(newJobParams)
                    self.log.debug("Production inherited from input", prod)
                else:
                    prod = job.getParam("Production").value
                    self.log.debug("Production:", f"{prod}")

                retVal = self.db.getProductionProcessingPassID(prod)
                if not retVal["OK"]:
                    return retVal

                res = self._getDataQuality(prod, runNumber)
                if not res["OK"]:
                    return res
                dqvalue = res["Value"]
                if not dqvalue:  # dqvalue can be None, if run/procid is not in newrunquality table
                    self.log.warn(
                        "Could not find run quality",
                        "for %d production (run number: %d)" % (int(prod), int(runNumber)),
                    )

            if len(tcks) > 1:
                self.log.debug("More than 1 TCK", f"[{','.join(tcks)}]")
                tck = -2
            elif len(tcks) == 1:
                tck = tcks.pop()
                self.log.debug("The output files of the job inherits the following TCK:", tck)
                if not job.exists("Tck"):
                    newJobParams = JobParameters()
                    newJobParams.name = "Tck"
                    newJobParams.value = tck
                    job.addJobParams(newJobParams)

        sumEventInputStat = 0
        sumEvtStat = 0
        sumLuminosity = 0

        if job.exists("JobType"):
            job.removeParam("JobType")

        inputfiles = job.inputFiles

        # This must be replaced by a single call!!!!
        # ## It is not urgent as we do not have a huge load on the database
        for i in inputfiles:
            fname = i.name
            res = self.db.getJobInfo(fname)
            if not res["OK"]:
                return res

            value = res["Value"]
            if value and value[0][2] is not None:
                sumEventInputStat += value[0][2]

            res = self.db.getFileMetadata([fname])
            if not res["OK"]:
                return res

            fileMetadata = res["Value"]["Successful"].get(fname)
            if fileMetadata:
                if fileMetadata["EventStat"] is not None:
                    sumEvtStat += fileMetadata["EventStat"]
                if fileMetadata["Luminosity"] is not None:
                    sumLuminosity += fileMetadata["Luminosity"]
                if dqvalue is None:
                    dqvalue = fileMetadata.get("DataqualityFlag", fileMetadata.get("DQFlag"))
            else:
                errMsg = f"Can not get the metadata of {fname} file"
                self.log.error(errMsg)
                return S_ERROR(errMsg)

        evtinput = 0
        if int(sumEvtStat) > int(sumEventInputStat):
            evtinput = sumEvtStat
        else:
            evtinput = sumEventInputStat

        if inputfiles:
            if not job.exists("EventInputStat"):
                newJobParams = JobParameters()
                newJobParams.name = "EventInputStat"
                newJobParams.value = str(evtinput)
                job.addJobParams(newJobParams)
            else:
                currentEventInputStat = job.getParam("EventInputStat")
                currentEventInputStat.value = evtinput

        self.log.debug("Luminosity:", sumLuminosity)
        outputFiles = job.outputFiles
        for outputfile in outputFiles:
            if outputfile.type not in ["LOG"] and sumLuminosity > 0 and not outputfile.exists("Luminosity"):
                newFileParams = FileParam()
                newFileParams.name = "Luminosity"
                newFileParams.value = sumLuminosity
                outputfile.addFileParam(newFileParams)
                self.log.debug("Luminosity added to ", outputfile.name)
            ################

        for param in job.parameters:
            if param.name == "RunNumber":
                value = int(param.value)
                if value <= 0 and len(job.inputFiles) == 0:
                    # The files which inherits the runs can be entered to the database
                    return S_ERROR("The run number not greater 0!")

        result = self.__insertJob(job)
        if not result["OK"]:
            config = job.configuration
            errorMessage = "Unable to create Job: {} , {}, {} .\n Error: {}".format(
                str(config.configName),
                str(config.configVersion),
                str(config.date),
                str(result["Message"]),
            )
            return S_ERROR(errorMessage)

        job.jobID = int(result["Value"])

        if job.exists("RunNumber"):  # case of, e.g., real data processing
            try:
                runnumber = int(job.getParam("RunNumber").value)
            except ValueError:
                runnumber = -1
            if runnumber != -1:
                self.log.verbose("Registering the run status for ", f"Run number {runnumber},  JobId {job.jobID}")
                result = self.db.insertRunStatus(runnumber, job.jobID, "N")
                if not result["OK"]:
                    errorMessage = ("Unable to register run status", runnumber + result["Message"])
                    self.log.error(errorMessage[0], errorMessage[1])
                    res = self.db.deleteJob(job.jobID)
                    if not res["OK"]:
                        self.log.warn("Unable to delete job", str(job.jobID) + res["Message"])
                    return S_ERROR(errorMessage[0])

                # we may be using HLT2 output to flag the runs: as a consequence we may have already flagged the run
                retVal = self._getDataQuality(runNumber=runnumber)
                if not retVal["OK"]:
                    return retVal
                if retVal["Value"]:  # if not "None", override what is found in the ancestors
                    dqvalue = retVal["Value"]
                    self.log.verbose("The run data quality flag for", "run %d is %s" % (runnumber, dqvalue))

            else:
                # we ran on multiple runs
                self.log.warn("Run number can not determined for production:", job.getParam("Production").value)

        inputFiles = job.inputFiles
        for inputfile in inputFiles:
            result = self.db.insertInputFile(job.jobID, inputfile.fileID)
            if not result["OK"]:
                errorMessage = ("Unable to insert input file", (str(inputfile.name)) + result["Message"])
                self.log.error(errorMessage[0], errorMessage[1])
                res = self.db.deleteJob(job.jobID)
                if not res["OK"]:
                    self.log.warn("Unable to delete job", str(job.jobID) + res["Message"])
                return S_ERROR(errorMessage[0])

        outputFiles = job.outputFiles
        prod = job.getParam("Production").value
        stepid = job.getParam("StepID").value
        retVal = self.db.getProductionOutputFileTypes(prod, stepid)
        if not retVal["OK"]:
            return retVal
        outputFileTypes = retVal["Value"]
        for outputfile in outputFiles:
            if dqvalue is not None:
                newFileParams = FileParam()
                newFileParams.name = "QualityId"
                newFileParams.value = dqvalue
                outputfile.addFileParam(newFileParams)
            elif not job.exists("RunNumber"):  # if it is MC
                newFileParams = FileParam()
                newFileParams.name = "QualityId"
                newFileParams.value = "OK"
                outputfile.addFileParam(newFileParams)
            ftype = outputfile.type
            if ftype in outputFileTypes:
                vFileParams = FileParam()
                vFileParams.name = "VisibilityFlag"
                vFileParams.value = outputFileTypes[ftype]
                outputfile.addFileParam(vFileParams)
                self.log.debug("The visibility flag is", outputFileTypes[ftype])

            result = self.__insertOutputFiles(job, outputfile)
            if not result["OK"]:
                errorMessage = (
                    "Unable to insert output file",
                    f"{outputfile.name} ! ERROR: {result['Message']}",
                )
                self.log.error(errorMessage[0], errorMessage[1])
                res = self.db.deleteInputFiles(job.jobID)
                if not res["OK"]:
                    self.log.warn("Unable to delete inputfiles of", str(job.jobID) + res["Message"])
                res = self.db.deleteJob(job.jobID)
                if not res["OK"]:
                    self.log.warn("Unable to delete job", str(job.jobID) + res["Message"])
                return S_ERROR(errorMessage[0])
            else:
                fileid = int(result["Value"])
                outputfile.fileID = fileid

            replicas = outputfile.replicas
            for replica in replicas:
                params = replica.params
                for param in params:
                    # just one param exist in params list, because JobReader only one param add to Replica
                    name = param.name
                result = self.db.updateReplicaRow(outputfile.fileID, "No")
                if not result["OK"]:
                    return S_ERROR(f"Unable to create Replica {str(name)} !")

        self.log.debug("End Processing!")

        return S_OK()

    @convertToReturnValue
    def _getRunNumbersAndTCKs(self, fileList):
        """Utility to get run numbers and TCKs of a list of files"""
        runnumbers = set()
        tcks = set()
        for lfn in fileList:
            for runtck in returnValueOrRaise(self.db.getRunNbAndTck(lfn)):
                if runtck[0]:
                    runnumbers.add(runtck[0])
                if runtck[1] and runtck[1] != "None":
                    tcks.add(runtck[1])
        return (runnumbers, tcks)

    @convertToReturnValue
    def _getDataQuality(self, prod=None, runNumber=None):
        """Return the DQ for a given prod and run number.
        A failure in finding the processing pass ID is a genuine error,
        while not finding a DQ flag can be okay (case of real data taking, not yet flagged)
        """
        if not runNumber:
            return None
        procID = returnValueOrRaise(self.db.getProductionProcessingPassID(prod or runNumber * -1))
        return self.db.getRunAndProcessingPassDataQuality(runNumber, procID).get("Value")

    def __insertJob(self, job):
        """Inserts the job to the database."""
        config = job.configuration

        production = None

        condParams = job.dataTakingCondition  # real data
        if condParams:
            datataking = condParams.parameters
            config = job.configuration

            ver = config.configVersion  # online bug fix
            ver = ver.capitalize()
            config.configVersion = ver
            self.log.debug("Data taking:", f"{datataking}")
            dtDescription = generateConditionDescription(datataking, config.configName)
            self.log.debug(dtDescription)
            datataking["Description"] = dtDescription

            res = self.db._getDataTakingConditionId(dtDescription)
            if not res["OK"]:
                self.log.error(
                    "Error retrieving the DataTaking Condition ID",
                    f"Description: {dtDescription}. Error: {res['Message']}",
                )
                return res

            daqid = res["Value"]
            # If there is no condition matching the description, create one
            if daqid == -1:  # yes... -1 for non existing conditions
                res = self.db.insertDataTakingCondDesc(dtDescription)
                if not res["OK"]:
                    self.log.error("Cannot insert DataTaking Condition Description", res["Message"])
                    return res
                daqid = res["Value"]

            # insert processing pass
            programName = None
            programVersion = None
            conddb = None
            dddb = None
            found = False
            for param in job.parameters:
                if param.name == "ProgramName":
                    programName = param.value
                elif param.name == "ProgramVersion":
                    programVersion = param.value
                elif param.name == "CondDB":
                    conddb = param.value
                elif param.name == "DDDB":
                    dddb = param.value
                elif param.name == "RunNumber":
                    production = int(param.value) * -1
                    found = True

            if job.exists("CondDB"):
                job.removeParam("CondDB")
            if job.exists("DDDB"):
                job.removeParam("DDDB")

            if not found:
                self.log.error("Run number is missing!")
                return S_ERROR("Run number is missing!")

            if "SmogInjection" in datataking:
                # "RunNumber" is checked several lines above (for production)
                gas = self.smogInjectionSMOG2Gas(datataking["SmogInjection"])
                if gas:
                    retVal = self.db.setSMOG2State(gas, False, [production * -1])
                    if not retVal["OK"]:
                        self.log.error(f"Can't set SMOG2 state {gas} for run {production * -1}, will ignore...")
                        # return S_ERROR("Can't set SMOG2 state for run") # TOCHANGE: Ignore till SMOG2 is checked

            retVal = self.db.getStepIdandNameForRUN(programName, programVersion, conddb, dddb)

            if not retVal["OK"]:
                return retVal

            stepid = retVal["Value"][0]

            # now we have to get the list of eventtypes
            eventtypes = []
            for outputFiles in job.outputFiles:
                for outPutfileParam in outputFiles.params:
                    outputFileParamName = outPutfileParam.name
                    if outputFileParamName == "EventTypeId":
                        eventtypes.append(int(outPutfileParam.value))

            steps = {
                "Steps": [
                    {
                        "StepId": stepid,
                        "StepName": retVal["Value"][1],
                        "ProcessingPass": retVal["Value"][1],
                        "Visible": "Y",
                        "OutputFileTypes": [{"FileType": "RAW"}],
                    }
                ]
            }

            self.log.debug("Pass_indexid", f"{steps}")
            self.log.debug("Data taking", f"{dtDescription}")
            self.log.debug("production", production)

            newJobParams = JobParameters()
            newJobParams.name = "StepID"
            newJobParams.value = str(stepid)
            job.addJobParams(newJobParams)

            message = f"StepID for run: {str(production)}"
            self.log.info(message, stepid)

            res = self.db.addProduction(
                production,
                simcond=None,
                daq=dtDescription,
                steps=steps["Steps"],
                inputproc="",
                configName=config.configName,
                configVersion=config.configVersion,
                eventType=eventtypes,
            )
            if res["OK"]:
                self.log.verbose("New processing pass has been created!")
                self.log.verbose("New production is:", production)
            elif job.exists("RunNumber"):
                self.log.warn("The run already registered!")
            else:
                self.log.error("Failing adding production", production + res["Message"])
                retVal = self.db.deleteStepContainer(production)
                if not retVal["OK"]:
                    return retVal
                return S_ERROR("Failing adding production")

        attrList = {"ConfigName": config.configName, "ConfigVersion": config.configVersion, "JobStart": None}

        for param in job.parameters:
            attrList[str(param.name)] = param.value

        res = self.db.checkProcessingPassAndSimCond(attrList["Production"])
        if not res["OK"]:
            self.log.error("check processing pass and simulation condition error", res["Message"])
        else:
            value = res["Value"]
            if value[0][0] == 0:
                errorMessage = "Missing processing pass and simulation conditions: "
                errorMessage += f"please fill it. Production = {str(attrList['Production'])}"
                self.log.warn(errorMessage)

        if attrList["JobStart"] is None:
            # date = config.date.split('-')
            # time = config.time.split(':')
            # dateAndTime = datetime.datetime(int(date[0]), int(date[1]), int(date[2]), int(time[0]), int(time[1]), 0, 0)
            attrList["JobStart"] = config.date + " " + config.time

        if production is not None:  # for the online registration
            attrList["Production"] = production

        res = self.db.insertJob(attrList)

        if not res["OK"] and production is not None and production < 0:
            self.log.error("Failed inserting job", res["Message"])
            retVal = self.db.deleteProductionsContainer(production)
            if not retVal["OK"]:
                self.log.error(retVal["Message"])
        return res

    #############################################################################
    def __insertOutputFiles(self, job, outputfile):
        """insert the files produced by a job."""
        attrList = {"FileName": outputfile.name, "FileTypeId": outputfile.typeID, "JobId": job.jobID}

        for param in outputfile.params:
            attrList[str(param.name)] = param.value
        return self.db.insertOutputFile(attrList)

    #############################################################################
    def processReplicas(self, replica):
        """process the replica registration request."""
        outputfile = replica.name
        self.log.debug("Processing replicas:", f"{outputfile}")
        fileID = -1

        delete = True

        replicaFileName = ""
        for param in replica.params:
            replicaFileName = param.file
            location = param.location
            delete = param.action == "Delete"

            result = self.db.checkfile(replicaFileName)
            if not result["OK"]:
                message = "No replica can be "
                if delete:
                    message += "removed"
                else:
                    message += "added"
                message += " to file " + str(replicaFileName) + " for " + str(location) + ".\n"
                return S_ERROR(message)
            else:
                fileID = int(result["Value"][0][0])
                self.log.debug("FileId:", fileID)

            if delete:
                result = DataManager().getReplicas(replicaFileName)
                replicaList = result["Value"]["Successful"]
                if len(replicaList) == 0:
                    result = self.db.updateReplicaRow(fileID, "No")
                    if not result["OK"]:
                        self.log.warn("Unable to set the Got_Replica flag for ", f"{replicaFileName}")
                        return S_ERROR("Unable to set the Got_Replica flag for ", f"{replicaFileName}")
            else:
                result = self.db.updateReplicaRow(fileID, "Yes")
                if not result["OK"]:
                    return S_ERROR("Unable to set the Got_Replica flag for " + str(replicaFileName))

        self.log.debug("End Processing replicas!")

        return S_OK()



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import shlex
import sys
import time
from asyncio.subprocess import create_subprocess_exec, PIPE, STDOUT
from tempfile import NamedTemporaryFile

from rich.markdown import Markdown
from rich.prompt import Prompt
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise

from .integrations import RepoIssue


async def do_checks(console, logbook, repo, request_ids):
    while True:
        choice = ask("Which request would you like to check?", choices=request_ids)
        issue_iid = int(request_ids[choice].split("#")[1].split("[")[0])
        request_id = int(choice)

        issue = RepoIssue(repo.get_issue(issue_iid))
        request_metadata = issue.metadata.requests[request_id]
        choice = ask("Which transformation would you like to check?", choices=request_metadata.transform_ids + ["all"])
        choice = request_metadata.transform_ids if choice == "all" else [int(choice)]
        for transform_id in choice:
            console.rule(f"Checking transformation {transform_id}")

            console.rule()
            choice = ask("Would you like to run the PM checks?", choices=["yes", "no"], default="yes")
            if choice == "yes":
                await pm_checks(console, logbook, issue, request_id, transform_id)
            else:
                console.print("Skipping PM checks")

            console.rule()
            choice = ask("Would you like to run the DM checks?", choices=["yes", "no"], default="yes")
            if choice == "yes":
                await dm_checks(console, logbook, issue, request_id, transform_id)
            else:
                console.print("Skipping DM checks")

        console.rule()
        choice = ask("Would you like to check if the buffer has been cleaned?", choices=["yes", "no"], default="yes")
        if choice == "yes":
            await buffer_check_removal(console, logbook, issue, request_id)
        else:
            console.print("Skipping buffer check")


async def pm_checks(console, logbook, issue, request_id, transform_id):
    console.print(f"Running PM checks for request {request_id} transformation {transform_id} ({issue.url})")
    cmd = ["dirac-production-check-descendants", str(transform_id), "--Force"]
    output = await run_command(cmd, console)

    choice = ask("Would you like to mark check as successful?", choices=["yes", "no"])
    if choice == "yes":
        issue.metadata.requests[request_id].checks.pm_check[transform_id] = True
        issue.update_metadata()
    maybe_post_logbook(
        console, logbook, issue.url, transform_id, choice == "yes", [(cmd, output)], ["Sprucing", "Production"]
    )


async def dm_checks(console, logbook, issue, request_id, transform_id):
    console.print(f"Running DM checks for request {request_id} transformation {transform_id} ({issue.url})")

    cmd = ["dirac-dms-check-fc2bkk", "--Production", transform_id]
    output = await run_command(cmd, console)

    cmd2 = ["dirac-dms-check-bkk2fc", "--Production", transform_id]
    output2 = await run_command(cmd2, console)

    choice = ask("Would you like to mark check as successful?", choices=["yes", "no"])
    if choice == "yes":
        issue.metadata.requests[request_id].checks.dm_check[transform_id] = True
        issue.update_metadata()
    maybe_post_logbook(
        console,
        logbook,
        issue.url,
        transform_id,
        choice == "yes",
        [(cmd, output), (cmd2, output2)],
        ["Sprucing", "Data Management", "Production"],
    )


async def run_command(cmd, console):
    cmd = list(map(str, cmd))
    console.rule()
    console.print(f"[bold]Output from:[/bold] {shlex.join(cmd)}")
    spinner = console.status(f"[bold]Running command:[/bold] {shlex.join(cmd)}", spinner="bouncingBall")
    start = time.monotonic()
    spinner.start()

    proc = await create_subprocess_exec(*cmd, stdout=PIPE, stderr=STDOUT)
    output = ""
    while line := await proc.stdout.readline():
        line = line.decode()
        output += line
        console.print(line, end="")
    await proc.communicate()
    spinner.stop()
    end = time.monotonic()
    console.print(f"[bold]Command finished with return code {proc.returncode} in {end - start:.2f}s[/bold]")
    console.rule()
    return output


def maybe_post_logbook(console, logbook, url, transform_id, success, cmds, systems):
    body = []
    if success:
        subject = "Checks successful"
        body.append("Checks have been successfully completed.")
    else:
        subject = "Checks failed"
        body.append("Checks were ran.")
    for cmd, output in cmds:
        body.append("")
        if cmd is not None:
            body.append("Command: " + shlex.join(map(str, cmd)))
            body.append("")
        body.append("Output:")
        body.append("```")
        body.append(output)
        body.append("```")

    body = "\n".join(body)

    console.print(Markdown(f"### Subject: {subject}\n\n{body}"))

    choice = ask("Would you like to make the above logbook entry?", choices=["yes", "no"])
    if choice == "yes":
        logbook.create_post(url, [transform_id], subject, body, {}, systems)


async def buffer_check_removal(console, logbook, issue, request_id):
    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.integrations import get_file_status

    transform_id = issue.metadata.requests[request_id].transform_ids[0]
    removal_id = issue.metadata.requests[request_id].removal

    body = []
    cmd_out = []
    spruce_file_status = {k["Status"]: v for k, v in await get_file_status(transform_id)}
    body.append(f"File statuses for first transformation {transform_id}: {spruce_file_status}")
    removal_file_status = {k["Status"]: v for k, v in await get_file_status(removal_id)}
    body.append(f"File statuses for removal transformation {removal_id}: {removal_file_status}")
    if not set(spruce_file_status).issubset({"Processed", "NotProcessed", "Removed"}):
        raise NotImplementedError()
    if not set(removal_file_status).issubset({"Processed", "Unused"}):
        raise NotImplementedError()
    not_processed_lfns = {
        x["LFN"]
        for x in returnValueOrRaise(
            TransformationClient().getTransformationFiles(
                {"TransformationID": transform_id, "Status": "NotProcessed"}, columns=["LFN"]
            )
        )
    }
    unused_lfns = {
        x["LFN"]
        for x in returnValueOrRaise(
            TransformationClient().getTransformationFiles(
                {"TransformationID": removal_id, "Status": "Unused"}, columns=["LFN"]
            )
        )
    }
    if lfns_to_clean := not_processed_lfns.intersection(unused_lfns):
        body.append(f"* {len(lfns_to_clean)} LFNs which are Unused due to being set NotProcessed:")
        body += lfns_to_clean
    if lfns := not_processed_lfns - unused_lfns:
        raise NotImplementedError(lfns)
    if lfns := unused_lfns - not_processed_lfns:
        raise NotImplementedError(lfns)
    body = "\n".join(body)
    cmd_out = [(None, body)]
    console.print(body)

    if lfns_to_clean:
        choice = ask(f"Would you like to remove {len(lfns_to_clean)} LFNs from the buffers?", choices=["yes", "no"])
        if choice == "yes":
            with NamedTemporaryFile("wt") as f:
                f.write("\n".join(lfns))
                f.flush()
                cmd = ["dirac-dms-remove-replicas", "--SE", "Tier1-Buffer", f"--File={f.name}"]
                output = await run_command(cmd, console)
                cmd_out.append((cmd, output))

    choice = ask("Would you like to mark check as successful?", choices=["yes", "no"])
    if choice == "yes":
        issue.metadata.requests[request_id].checks.dm_clean[request_id] = True
        issue.update_metadata()
    maybe_post_logbook(
        console,
        logbook,
        issue.url,
        transform_id,
        choice == "yes",
        cmd_out,
        ["Sprucing", "Data Management", "Production"],
    )


def ask(question, choices, **kwargs):
    choices = sorted(str(choices) for choices in choices)
    choice = Prompt.ask(question, choices=choices + ["quit"], **kwargs)
    if choice == "quit":
        sys.exit(0)
    return choice



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

from datetime import datetime, timedelta

import pytest
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError

from LHCbDIRAC.ProductionManagementSystem.DB.AnalysisProductionsDB import AnalysisProductionsDB
from LHCbDIRAC.ProductionManagementSystem.DB.AnalysisProductionsObjects import AnalysisSample as AP


@pytest.fixture
def apdb():
    db = AnalysisProductionsDB(url="sqlite+pysqlite:///:memory:")
    yield db


REQUEST_1 = {
    "request_id": 1234,
    "filetype": "TUPLE.ROOT",
    "name": "MySample",
    "version": "v1r2p3",
    "wg": "MyWG",
    "analysis": "MyAnalysis",
    "extra_info": {
        "transformations": [],
        "merge_request": "https://gitlab.cern.ch/lhcb-datapkg/AnalysisProductions/-/merge_requests/0",
        "jira_task": "https://its.cern.ch/jira/browse/WGP-0",
    },
    "validity_start": datetime.now() - timedelta(days=1),
    "housekeeping_interaction_due": None,
    "owners": [],
    "auto_tags": [
        {"name": "config", "value": "MC"},
        {"name": "polarity", "value": "MagDown"},
        {"name": "eventtype", "value": "23133002"},
        {"name": "datatype", "value": "2012"},
    ],
    "publications": [],
}

REQUEST_2 = {
    "request_id": 987,
    "filetype": "TUPLE.ROOT",
    "name": "AnotherSample",
    "version": "v1r2p4",
    "wg": "AnotherWG",
    "analysis": "AnotherAnalysis",
    "extra_info": {
        "transformations": [],
        "merge_request": "https://gitlab.cern.ch/lhcb-datapkg/AnalysisProductions/-/merge_requests/0",
        "jira_task": "https://its.cern.ch/jira/browse/WGP-0",
    },
    "validity_start": datetime.now() - timedelta(days=1),
    "housekeeping_interaction_due": None,
    "owners": [],
    "auto_tags": [
        {"name": "config", "value": "LHCb"},
        {"name": "polarity", "value": "MagUp"},
        {"name": "datatype", "value": "2018"},
    ],
    "publications": [],
}

REQUEST_3 = {
    "request_id": 988,
    "filetype": "TUPLE.ROOT",
    "name": "AnotherSample",
    "version": "v1r2p3",
    "wg": "MyWG",
    "analysis": "MyAnalysis",
    "extra_info": {
        "transformations": [],
        "merge_request": "https://gitlab.cern.ch/lhcb-datapkg/AnalysisProductions/-/merge_requests/0",
        "jira_task": "https://its.cern.ch/jira/browse/WGP-0",
    },
    "validity_start": datetime.now() - timedelta(days=1),
    "housekeeping_interaction_due": None,
    "owners": [],
    "auto_tags": [
        {"name": "config", "value": "LHCb"},
        {"name": "polarity", "value": "MagUp"},
        {"name": "datatype", "value": "2018"},
    ],
    "publications": [],
}

TRANSFORMS_1a = {
    "id": 45,
    "status": "Archived",
    "steps": [
        {
            "stepID": 50,
            "application": "DaVinci/v45r6",
            "extras": ["AnalysisProductions.v0r0p2510752", "ProdConf"],
            "options": ["$SOMETHING/a.py"],
        }
    ],
    "used": False,
}

TRANSFORMS_1b = {
    "id": 47,
    "status": "Archived",
    "steps": [
        {
            "stepID": 51,
            "application": "Noether/v1r4",
            "extras": ["AppConfig.v3r398", "ProdConf"],
            "options": ["$SOMETHING/b.py"],
        }
    ],
    "used": True,
}


def _compareRequest(orig, new):
    assert new["sample_id"] is not None
    assert orig["request_id"] == new["request_id"]
    assert new["state"] is not None

    assert orig["wg"].lower() == new["wg"]
    assert orig["analysis"].lower() == new["analysis"]
    assert orig["version"].lower() == new["version"]
    assert orig["name"].lower() == new["name"]
    assert orig["extra_info"]["jira_task"] == new["jira_task"]
    assert orig["extra_info"]["merge_request"] == new["merge_request"]

    assert orig["owners"] == []
    assert new["owners"] == []
    assert orig["extra_info"]["transformations"] == new["transformations"]

    assert orig["validity_start"] == new["validity_start"]
    assert new["validity_end"] is None
    assert new["last_state_update"] is not None


def test_empty(apdb):
    assert apdb.listAnalyses() == {}
    assert apdb.listAnalyses(at_time=None) == {}
    assert apdb.listAnalyses(at_time=datetime.now()) == {}

    assert apdb.getKnownAutoTags() == set()

    assert apdb.getProductions() == []
    assert apdb.getProductions(wg="MyWG", analysis="MyAnalysis") == []
    assert apdb.getProductions(at_time=None) == []
    assert apdb.getProductions(at_time=datetime.now()) == []

    assert apdb.getArchivedRequests() == []
    assert apdb.getArchivedRequests(state="waiting") == []
    assert apdb.getArchivedRequests(state="ready") == []


def test_ownership(apdb):
    assert apdb.getOwners(wg="mywg", analysis="myanalysis") == []
    apdb.setOwners(wg="mywg", analysis="myanalysis", owners=["auser"])
    assert apdb.getOwners(wg="mywg", analysis="myanalysis") == ["auser"]

    assert apdb.getOwners(wg="mywg2", analysis="myanalysis") == []
    apdb.setOwners(wg="mywg2", analysis="myanalysis", owners=["user2"])
    assert apdb.getOwners(wg="mywg2", analysis="myanalysis") == ["user2"]

    assert apdb.getOwners(wg="mywg", analysis="myanalysis2") == []
    apdb.setOwners(wg="mywg", analysis="myanalysis2", owners=["user3"])
    assert apdb.getOwners(wg="mywg", analysis="myanalysis2") == ["user3"]

    apdb.setOwners(wg="mywg", analysis="myanalysis2", owners=[])
    assert apdb.getOwners(wg="mywg", analysis="myanalysis2") == []

    apdb.setOwners(wg="mywg", analysis="myanalysis", owners=["user4", "user5"])
    assert apdb.getOwners(wg="mywg", analysis="myanalysis") == ["user4", "user5"]

    apdb.setOwners(wg="mywg", analysis="myanalysis", owners=[])
    assert apdb.getOwners(wg="mywg", analysis="myanalysis") == []


def test_registerNew(apdb):
    newRequests = apdb.registerRequests([REQUEST_1])
    assert len(newRequests) == 1

    newRequest = newRequests[0]
    assert newRequest["sample_id"] == 1
    assert newRequest["state"] == "waiting"
    _compareRequest(REQUEST_1, newRequest)

    assert apdb.listAnalyses() == {"mywg": ["myanalysis"]}

    assert apdb.listAnalyses2() == [
        {
            "analysis": "myanalysis",
            "n_active": 0,
            "n_ready": 0,
            "n_replicating": 0,
            "n_total": 1,
            "n_waiting": 1,
            "owners": [],
            "wg": "mywg",
            "earliest_housekeeping_due": (
                datetime.today().replace(hour=0, minute=0, second=0, microsecond=0) + timedelta(days=90)
            ),
        }
    ]

    assert apdb.getProductions() == [newRequest]
    assert apdb.getProductions(wg="mywg", analysis="myanalysis") == [newRequest]
    assert apdb.getProductions(wg="MYWG", analysis="MYANALYSIS") == [newRequest]
    assert apdb.getProductions(name="mysample") == [newRequest]
    assert apdb.getProductions(name="MySample") == [newRequest]
    assert apdb.getProductions(name="anothersample") == []
    assert apdb.getProductions(state="waiting") == [newRequest]
    assert apdb.getProductions(state="active") == []
    assert apdb.getProductions(version="v1r2p3") == [newRequest]
    assert apdb.getProductions(version="v1r2p4") == []

    assert apdb.getKnownAutoTags() == {"config", "polarity", "eventtype", "datatype"}
    assert apdb.getTags("MyWG", "MyAnalysis") == {
        1: {
            "config": "mc",
            "polarity": "magdown",
            "eventtype": "23133002",
            "datatype": "2012",
        }
    }


def test_duplicate(apdb):
    apdb.registerRequests([REQUEST_1])
    with pytest.raises(ValueError, match=f"Already registered.*{REQUEST_1['request_id']}.*"):
        apdb.registerRequests([REQUEST_1])


def test_listRequests(apdb):
    assert apdb.listRequests() == []
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    requests = apdb.listRequests()
    assert len(requests) == 3
    for request in requests:
        assert request["tags"] == []
        assert isinstance(request["autotags"], dict)
        if "eventtype" in request["autotags"]:
            assert {"config", "datatype", "polarity", "eventtype"} == set(request["autotags"])
        else:
            assert {"config", "datatype", "polarity"} == set(request["autotags"])

    apdb.archiveSamples([apdb.getProductions()[0]["sample_id"]])

    requests = apdb.listRequests()
    assert len(requests) == 2


def test_multiple_samples_per_request(apdb: AnalysisProductionsDB):
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    assert {"myanalysis", "anotheranalysis"} == {x["analysis"] for x in apdb.listAnalyses2()}
    orig1 = apdb.getProductions(wg=REQUEST_1["wg"], analysis=REQUEST_1["analysis"], name=REQUEST_1["name"])
    assert len(orig1) == 1

    apdb.addRequestsToAnalysis("newwg", "newanalysis", [(REQUEST_1["request_id"], "TUPLE.ROOT")])
    assert {"newanalysis", "myanalysis", "anotheranalysis"} == {x["analysis"] for x in apdb.listAnalyses2()}

    with pytest.raises(ValueError, match="Some requests are already registered"):
        apdb.addRequestsToAnalysis(
            "newwg", "newanalysis", [(REQUEST_1["request_id"], "TUPLE.ROOT"), (REQUEST_2["request_id"], "TUPLE.ROOT")]
        )
    assert len(apdb.getProductions(wg="newwg", analysis="newanalysis")) == 1

    apdb.addRequestsToAnalysis(
        "newwg", "newanalysis", [(REQUEST_2["request_id"], "TUPLE.ROOT"), (REQUEST_3["request_id"], "TUPLE.ROOT")]
    )
    assert len(apdb.getProductions(wg="newwg", analysis="newanalysis")) == 3

    ref1 = apdb.getProductions(wg="newwg", analysis="newanalysis", name=REQUEST_1["name"])
    assert len(ref1) == 1
    for key in set(orig1[0]) | set(ref1[0]):
        if key not in {"validity_start", "wg", "analysis", "sample_id"}:
            assert orig1[0][key] == ref1[0][key], key


def test_archival(apdb):
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])

    assert apdb.getArchivedRequests() == []
    assert apdb.getArchivedRequests(state="waiting") == []
    assert apdb.getArchivedRequests(state="ready") == []

    apdb.archiveSamples([1])

    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 1
    assert archivedRequests[0]["request_id"] == 1234
    archivedRequests = apdb.getArchivedRequests(state="waiting")
    assert len(archivedRequests) == 1
    assert archivedRequests[0]["request_id"] == 1234
    assert apdb.getArchivedRequests(state="ready") == []

    with pytest.raises(ValueError, match=r".*have already been archived.*"):
        apdb.archiveSamples([2, 1, 3])
    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 1
    assert archivedRequests[0]["request_id"] == 1234

    with pytest.raises(ValueError, match=r"Unknown sample IDs passed.*"):
        apdb.archiveSamples([2, 12, 3])
    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 1
    assert archivedRequests[0]["request_id"] == 1234

    apdb.archiveSamples([2, 3])
    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 3


def test_archiveAtTime(apdb):
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])

    assert apdb.getArchivedRequests() == []
    assert apdb.getArchivedRequests(state="waiting") == []
    assert apdb.getArchivedRequests(state="ready") == []

    # archive in the past
    archive_time_past = datetime.now() - timedelta(days=1)
    apdb.archiveSamplesAtSpecificTime([1], archive_time_past)

    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 1
    assert archivedRequests[0]["request_id"] == 1234
    archivedRequests = apdb.getArchivedRequests(state="waiting")
    assert archivedRequests[0]["request_id"] == 1234
    assert apdb.getArchivedRequests(state="ready") == []

    with pytest.raises(ValueError, match=r".*have already been archived.*"):
        apdb.archiveSamplesAtSpecificTime([2, 1, 3], archive_time_past)
    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 1
    assert archivedRequests[0]["request_id"] == 1234

    with pytest.raises(ValueError, match=r"Unknown sample IDs passed.*"):
        apdb.archiveSamplesAtSpecificTime([2, 12, 3], archive_time_past)
    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 1
    assert archivedRequests[0]["request_id"] == 1234

    # archive in the future
    archive_time_future = datetime.now() + timedelta(days=1)
    apdb.archiveSamplesAtSpecificTime([2, 3], archive_time_future)
    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 1


def test_addPublication(apdb):
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    requests = apdb.listRequests()
    assert len(requests) == 3

    def result_has_pub(r, n, sample_id=None):
        if n not in r:
            return False
        if sample_id:
            if not any(i["sample_id"] == sample_id for i in r[n]):
                return False

        return True

    sample_ids = [1, 2, 3]

    prods_with_pubs = apdb.getProductions(require_has_publication=True)
    assert (
        len(prods_with_pubs) == 0
    ), "getProductions(require_has_publication=True) did not return empty, when it should have"

    with pytest.raises(ValueError, match=r"This publication number is too long.*"):
        apdb.addPublication(sample_ids, "LHCb-" + "h" * 75)

    apdb.addPublication(sample_ids, "LHCb-ANA-YYYY-NNN")

    sample_publications = apdb.getPublications(sample_ids)
    assert result_has_pub(sample_publications, "LHCb-ANA-YYYY-NNN"), 'Result doesn\'t have "LHCb-ANA-YYYY-NNN" at all'
    assert result_has_pub(
        sample_publications, "LHCb-ANA-YYYY-NNN", sample_id=sample_ids[0]
    ), f'Result doesn\'t have "LHCb-ANA-YYYY-NNN" for sample_id {sample_ids[0]}'
    assert result_has_pub(
        sample_publications, "LHCb-ANA-YYYY-NNN", sample_id=sample_ids[1]
    ), f'Result doesn\'t have "LHCb-ANA-YYYY-NNN" for sample_id {sample_ids[1]}'
    assert result_has_pub(
        sample_publications, "LHCb-ANA-YYYY-NNN", sample_id=sample_ids[2]
    ), f'Result doesn\'t have "LHCb-ANA-YYYY-NNN" for sample_id {sample_ids[2]}'

    with pytest.raises(IntegrityError):
        apdb.addPublication(sample_ids, "LHCb-ANA-YYYY-NNN")

    apdb.addPublication([sample_ids[0]], "LHCb-PAPER-YYYY-NNN")
    sample_publications = apdb.getPublications(sample_ids)

    assert result_has_pub(
        sample_publications, "LHCb-PAPER-YYYY-NNN"
    ), 'Result doesn\'t have "LHCb-PAPER-YYYY-NNN" at all'
    assert result_has_pub(
        sample_publications, "LHCb-ANA-YYYY-NNN", sample_id=sample_ids[0]
    ), f'Result doesn\'t have "LHCb-ANA-YYYY-NNN" for sample_id {sample_ids[0]}'
    assert result_has_pub(
        sample_publications, "LHCb-PAPER-YYYY-NNN", sample_id=sample_ids[0]
    ), f'Result doesn\'t have "LHCb-PAPER-YYYY-NNN" for sample_id {sample_ids[0]}'
    assert not result_has_pub(
        sample_publications, "LHCb-PAPER-YYYY-NNN", sample_id=sample_ids[1]
    ), f'Result shouldn\'t have "LHCb-PAPER-YYYY-NNN" for sample_id {sample_ids[1]}'
    assert not result_has_pub(
        sample_publications, "LHCb-PAPER-YYYY-NNN", sample_id=sample_ids[2]
    ), f'Result shouldn\'t have "LHCb-PAPER-YYYY-NNN" for sample_id {sample_ids[2]}'

    prods_with_pubs = apdb.getProductions()

    assert len(prods_with_pubs) == 3

    assert "LHCb-ANA-YYYY-NNN" in prods_with_pubs[0]["publications"]
    assert "LHCb-ANA-YYYY-NNN" in prods_with_pubs[1]["publications"]
    assert "LHCb-ANA-YYYY-NNN" in prods_with_pubs[2]["publications"]
    assert "LHCb-PAPER-YYYY-NNN" in prods_with_pubs[0]["publications"]
    assert "LHCb-PAPER-YYYY-NNN" not in prods_with_pubs[1]["publications"]
    assert "LHCb-PAPER-YYYY-NNN" not in prods_with_pubs[2]["publications"]

    prods_with_pubs = apdb.getProductions(require_has_publication=True)
    assert (
        len(prods_with_pubs) == 3
    ), "getProductions(require_has_publication=True) did not return the three requests with publications"


def test_get_and_delayHousekeepingInteractionDue(apdb):
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    requests = apdb.listRequests()
    assert len(requests) == 3

    sample_ids = [1, 2, 3]

    assert apdb.getArchivedRequests() == []

    def getHousekeepingInteractionDue(sample_ids: list[int]):
        query = select(AP.sample_id, AP.housekeeping_interaction_due)
        query = query.filter(AP.sample_id.in_(sample_ids))
        query = query.filter(AP.validity_end.is_(None))

        # this returns due dates for existing and non-archived requests
        # non-existing requests are omitted
        # archived requests are omitted
        # existing requests with no due date are returned but with a null value
        with apdb.session as session:
            return {i: d for i, d in session.execute(query).all()}

    # check that house keeping dates are returned at all even if None
    # if they don't exist, they won't be returned
    hkdates = getHousekeepingInteractionDue(sample_ids)
    assert all(sample_id in hkdates.keys() for sample_id in sample_ids)
    assert len(hkdates) == 3

    due_datetime_200d = datetime.now() + timedelta(days=200)
    due_datetime_1y = datetime.now() + timedelta(days=365)
    # delay a non-existent request
    with pytest.raises(ValueError, match=r"Unknown sample IDs passed.*"):
        apdb.delayHousekeepingInteractionDue(sample_ids + [999999999, 990999], due_datetime_200d)
    hkdates = getHousekeepingInteractionDue(sample_ids)

    # check nothing changed
    assert all(hkdates[sample_id] != due_datetime_200d for sample_id in sample_ids)

    # set housekeeping 200 days in the future for one request
    apdb.delayHousekeepingInteractionDue([sample_ids[0]], due_datetime_200d)
    hkdates = getHousekeepingInteractionDue(sample_ids)
    assert hkdates[sample_ids[0]] == due_datetime_200d
    assert all(sample_id in hkdates.keys() for sample_id in sample_ids)

    # archive one of the requests
    apdb.archiveSamples([sample_ids[0]])

    archivedRequests = apdb.getArchivedRequests()
    assert len(archivedRequests) == 1

    # shouldn't be returned anymore
    hkdates = getHousekeepingInteractionDue(sample_ids)
    assert len(hkdates.values()) == 2

    with pytest.raises(ValueError, match=r"Some samples have already been archived*"):
        apdb.delayHousekeepingInteractionDue(sample_ids, due_datetime_200d)

    hkdates = getHousekeepingInteractionDue(sample_ids)
    # check nothing changed
    assert all(hkdates[sample_id] != due_datetime_200d for sample_id in sample_ids[1:])
    assert len(hkdates) == 2

    # set the dates for the rest of these
    apdb.delayHousekeepingInteractionDue(sample_ids[1:], due_datetime_1y)

    # check everything changed as we expect
    hkdates = getHousekeepingInteractionDue(sample_ids)
    assert all(
        hkdates[sample_id] != due_datetime_200d and hkdates[sample_id] == due_datetime_1y
        for sample_id in sample_ids[1:]
    )
    assert len(hkdates) == 2

    due_now = apdb.getHousekeepingInteractionDueNow()
    assert len(due_now) == 0

    # set housekeeping in the past and check it is marked as due
    due_datetime_past = datetime.now() - timedelta(days=5)
    apdb.delayHousekeepingInteractionDue([sample_ids[-1]], due_datetime_past)
    due_now = apdb.getHousekeepingInteractionDueNow()
    assert len(due_now) == 1

    # set analysis owner and check it turns up.
    wg_o = due_now[0]["wg"]
    ana_o = due_now[0]["analysis"]
    apdb.setOwners(wg=wg_o, analysis=ana_o, owners=["user_who_forgot", "and_another"])

    due_now = apdb.getHousekeepingInteractionDueNow()
    assert "user_who_forgot" in due_now[0]["owner_usernames"]


def test_registerTransformations(apdb):
    apdb.registerRequests([REQUEST_1])

    with pytest.raises(ValueError):
        apdb.registerTransformations({})

    apdb.registerTransformations({1234: {"TUPLE.ROOT": [TRANSFORMS_1a]}})
    assert apdb.getProductions()[0]["transformations"] == [TRANSFORMS_1a]

    with pytest.raises(ValueError, match=r".*already known.*"):
        apdb.registerTransformations({1234: {"TUPLE.ROOT": [TRANSFORMS_1a, TRANSFORMS_1b]}})
    assert apdb.getProductions()[0]["transformations"] == [TRANSFORMS_1a]

    apdb.registerTransformations({1234: {"TUPLE.ROOT": [TRANSFORMS_1b]}})
    assert apdb.getProductions()[0]["transformations"] == [TRANSFORMS_1a, TRANSFORMS_1b]

    apdb.deregisterTransformations({1234: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]}})
    assert apdb.getProductions()[0]["transformations"] == [TRANSFORMS_1b]


def test_registerTransformationsError(apdb):
    """Ensure that nothing is changed if an invalid request ID is passed."""
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    with pytest.raises(ValueError, match=r"Did not find requests for IDs: \[\(99999, 'TUPLE.ROOT'\)\]"):
        apdb.registerTransformations(
            {
                987: {"TUPLE.ROOT": [TRANSFORMS_1a]},
                99999: {"TUPLE.ROOT": [TRANSFORMS_1a]},
                988: {"TUPLE.ROOT": [TRANSFORMS_1a]},
            }
        )
    for prod in apdb.getProductions():
        assert prod["transformations"] == [], prod


def test_deregisterTransformationsError(apdb):
    """Ensure that nothing is changed if an invalid request ID is passed."""
    apdb.registerRequests([REQUEST_2, REQUEST_3])
    apdb.registerTransformations({987: {"TUPLE.ROOT": [TRANSFORMS_1a]}, 988: {"TUPLE.ROOT": [TRANSFORMS_1a]}})
    with pytest.raises(ValueError, match=r"Did not find requests for IDs: \[\(99999, \'TUPLE.ROOT\'\)\]"):
        apdb.deregisterTransformations(
            {
                987: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]},
                99999: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]},
                988: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]},
            }
        )
    for prod in apdb.getProductions():
        assert prod["transformations"] == [TRANSFORMS_1a], prod


def test_deregisterTransformationsError2(apdb):
    """Ensure that nothing is changed if an invalid request ID is passed."""
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    apdb.registerTransformations(
        {
            1234: {"TUPLE.ROOT": [TRANSFORMS_1a]},
            987: {"TUPLE.ROOT": [TRANSFORMS_1a]},
            988: {"TUPLE.ROOT": [TRANSFORMS_1a]},
        }
    )
    with pytest.raises(ValueError, match=r"Transformation 47 is not known"):
        apdb.deregisterTransformations(
            {
                1234: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]},
                987: {"TUPLE.ROOT": [TRANSFORMS_1a["id"], 47]},
                988: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]},
            }
        )
    for prod in apdb.getProductions():
        assert prod["transformations"] == [TRANSFORMS_1a], prod

    apdb.deregisterTransformations(
        {
            1234: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]},
            987: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]},
            988: {"TUPLE.ROOT": [TRANSFORMS_1a["id"]]},
        }
    )
    for prod in apdb.getProductions():
        assert prod["transformations"] == [], prod


def test_setState(apdb):
    apdb.registerRequests([REQUEST_1])
    prod1 = apdb.getProductions()[0]
    assert prod1["state"] == "waiting"
    assert "progress" not in prod1

    apdb.setState({(1234, "TUPLE.ROOT"): {"state": "active", "progress": 0.5}})

    prod2 = apdb.getProductions()[0]
    assert prod1["last_state_update"] <= prod2["last_state_update"]
    assert prod2["state"] == "active"
    assert prod2["progress"] == 0.5

    apdb.setState({(1234, "TUPLE.ROOT"): {"state": "ready", "progress": None}})

    prod3 = apdb.getProductions()[0]
    assert prod2["last_state_update"] <= prod3["last_state_update"]
    assert prod3["state"] == "ready"
    assert "progress" not in prod3


def test_setStateMultiple(apdb):
    apdb.registerRequests([REQUEST_1, REQUEST_3])
    apdb.setState(
        {
            (1234, "TUPLE.ROOT"): {"state": "active", "progress": 0.5},
            (988, "TUPLE.ROOT"): {"state": "active", "progress": 0.4},
        }
    )
    prod1, prod2 = apdb.getProductions()
    assert prod1["request_id"] == 1234
    assert prod1["state"] == "active"
    assert prod1["progress"] == 0.5
    assert prod2["request_id"] == 988
    assert prod2["state"] == "active"
    assert prod2["progress"] == 0.4
    with pytest.raises(ValueError, match=r"Failed to update Request.*"):
        apdb.setState(
            {
                (1234, "TUPLE.ROOT"): {"state": "active", "progress": 0.6},
                (99999, "TUPLE.ROOT"): {"state": "active", "progress": 0.3},
                (988, "TUPLE.ROOT"): {"state": "active", "progress": 0.7},
            }
        )
    prod1, prod2 = apdb.getProductions()
    assert prod1["request_id"] == 1234
    assert prod1["state"] == "active"
    assert prod1["progress"] == 0.5
    assert prod2["request_id"] == 988
    assert prod2["state"] == "active"
    assert prod2["progress"] == 0.4


def test_getKnownAutoTags(apdb):
    apdb.registerRequests([REQUEST_2])
    assert set(apdb.getKnownAutoTags()) == {"config", "polarity", "datatype"}
    apdb.registerRequests([REQUEST_1])
    assert set(apdb.getKnownAutoTags()) == {"config", "polarity", "eventtype", "datatype"}


def test_setTagsNoChanges(apdb):
    """Setting tags to the same value should be a no-op."""
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    apdb.setTags(
        {
            1: {"config": "mc", "polarity": "magdown", "eventtype": "23133002", "datatype": "2012"},
            2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
        },
        {
            1: {"config": "mc", "polarity": "magdown", "eventtype": "23133002", "datatype": "2012"},
            2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
        },
    )
    assert apdb.getTags("MyWG", "MyAnalysis") == {
        1: {"config": "mc", "polarity": "magdown", "eventtype": "23133002", "datatype": "2012"},
        3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
    }
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}
    }


def test_setTagsCaseInsensitive(apdb):
    """Tags should always be lowercased"""
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    apdb.setTags(
        {
            2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
        },
        {
            2: {"config": "lhcb", "Polarity": "MagUp", "datatype": "2018"},
        },
    )
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}
    }
    apdb.setTags(
        {
            2: {"config": "lhcb", "Polarity": "MagUp", "datatype": "2018"},
        },
        {
            2: {"config": "lhcb", "polarity": "MagUp", "datatype": "2018"},
        },
    )
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}
    }
    apdb.setTags(
        {
            2: {"config": "lhcb", "Polarity": "MagUp", "datatype": "2018"},
        },
        {
            2: {"config": "lhcb", "polarity": "MagUp", "dataType": "2018", "HELLO": "WORLD", "FoO": 123},
        },
    )
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "hello": "world", "foo": "123"}
    }
    with pytest.raises(ValueError, match=r".*contains duplicate keys.*"):
        apdb.setTags(
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
            },
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "hello": "world", "HELLO": 123},
            },
        )
    assert apdb.getTags("MyWG", "MyAnalysis") == {
        1: {"config": "mc", "polarity": "magdown", "eventtype": "23133002", "datatype": "2012"},
        3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
    }


def test_setTagsWithChanges(apdb):
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    apdb.setTags(
        {
            3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
        },
        {
            3: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "DpToKpipi_SS"},
        },
    )
    assert apdb.getTags("MyWG", "MyAnalysis") == {
        1: {"config": "mc", "polarity": "magdown", "eventtype": "23133002", "datatype": "2012"},
        3: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "dptokpipi_ss"},
    }
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}
    }


def test_setTagsMissingOldTags(apdb):
    """Tags cannot be changed if oldTags isn't up to date"""
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    with pytest.raises(ValueError, match=r".*must contain the same keys.*"):
        apdb.setTags(
            {},
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "DpToKpipi_SS"},
            },
        )
    assert apdb.getTags("MyWG", "MyAnalysis") == {
        1: {"config": "mc", "polarity": "magdown", "eventtype": "23133002", "datatype": "2012"},
        3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
    }
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}
    }


def test_setTagsWrongOldTags(apdb):
    """Tags cannot be changed if oldTags isn't up to date"""
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    apdb.setTags(
        {2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}},
        {2: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "DpToKpipi_SS"}},
    )
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "dptokpipi_ss"}
    }
    with pytest.raises(ValueError, match=r"oldTags is out of date"):
        apdb.setTags(
            {2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}},
            {2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}},
        )
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "dptokpipi_ss"}
    }
    with pytest.raises(ValueError, match=r"oldTags is out of date"):
        apdb.setTags(
            {2: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "dptokpipi_os"}},
            {2: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "dptokpipi_ws"}},
        )
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "dptokpipi_ss"}
    }


# def test_setTagsOutdated(apdb):


def test_setTagsAutoTags(apdb):
    """Ensure AutoTags cannot be added, removed or modified"""
    apdb.registerRequests([REQUEST_1, REQUEST_2, REQUEST_3])
    with pytest.raises(ValueError, match=r"Cannot modify AutoTags.*"):
        apdb.setTags(
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
                2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
            },
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "DpToKpipi_SS"},
                2: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "eventtype": "90000000"},
            },
        )
    with pytest.raises(ValueError, match=r"Cannot modify AutoTags.*"):
        apdb.setTags(
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
                2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
            },
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "DpToKpipi_SS"},
                2: {"config": "lhcb", "polarity": "magdown", "datatype": "2018"},
            },
        )
    with pytest.raises(ValueError, match=r"Cannot modify AutoTags.*"):
        apdb.setTags(
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
                2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
            },
            {
                3: {"config": "lhcb", "polarity": "magup", "datatype": "2018", "sample": "DpToKpipi_SS"},
                2: {"config": "lhcb", "polarity": "magup"},
            },
        )
    assert apdb.getTags("MyWG", "MyAnalysis") == {
        1: {"config": "mc", "polarity": "magdown", "eventtype": "23133002", "datatype": "2012"},
        3: {"config": "lhcb", "polarity": "magup", "datatype": "2018"},
    }
    assert apdb.getTags("AnotherWG", "AnotherAnalysis") == {
        2: {"config": "lhcb", "polarity": "magup", "datatype": "2018"}
    }



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve production summary from the Bookkeeping."""
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()
    Script.setUsageMessage(__doc__ + "\n".join(["Usage:", f"  {Script.scriptName} [option|cfgfile] ..."]))
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()

    bkQuery = dmScript.getBKQuery()
    if not bkQuery:
        gLogger.error("No BKQuery given...")
        DIRAC.exit(1)

    bkQueryDict = bkQuery.getQueryDict()
    dictItems = (
        "ConfigName",
        "ConfigVersion",
        "Production",
        "ConditionDescription",
        "ProcessingPass",
        "FileType",
        "EventType",
    )
    for item in dictItems:
        bkQueryDict.setdefault(item, "ALL")
    for item in list(bkQueryDict):
        if item not in dictItems:
            bkQueryDict.pop(item)

    gLogger.verbose("BKQuery:", bkQueryDict)
    res = bk.getProductionSummary(bkQueryDict)

    if not res["OK"]:
        gLogger.error(res["Message"])
        DIRAC.exit(1)

    records = res["Value"]["Records"]
    params = res["Value"]["ParameterNames"]
    width = 20

    gLogger.showHeaders(False)

    gLogger.notice("")
    gLogger.notice(
        params[0].ljust(30)
        + str(params[1]).ljust(30)
        + str(params[2]).ljust(30)
        + str(params[3]).ljust(30)
        + str(params[4]).ljust(30)
        + str(params[5]).ljust(30)
        + str(params[6]).ljust(20)
        + str(params[7]).ljust(20)
        + str(params[8]).ljust(20),
    )
    gLogger.notice("")
    for record in records:
        gLogger.notice(
            str(record[0]).ljust(15)
            + str(record[1]).ljust(15)
            + str(record[2]).ljust(20)
            + str(record[3]).ljust(width)
            + str(record[4]).ljust(width)
            + str(record[5]).ljust(width)
            + str(record[6]).ljust(width)
            + str(record[7]).ljust(width)
            + str(record[8]).ljust(width),
        )

    gLogger.notice("")
    gLogger.notice("TotalRecords = %d" % res["Value"]["TotalRecords"])


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Start or stop the production(s)"""
import DIRAC
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} <Production ID> |<Production ID>",
                "Arguments:",
                "  <Production ID>:      DIRAC Production Id",
            ]
        )
    )

    Script.registerSwitch("t", "start", "Start the production")
    Script.registerSwitch("p", "stop", "Stop the production")

    Script.parseCommandLine(ignoreErrors=True)

    args = Script.getPositionalArgs()
    if len(args) < 1:
        Script.showHelp(exitCode=2)

    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    diracProd = DiracProduction()
    exitCode = 0
    errorList = []
    start = False
    stop = False
    action = ""

    switches = Script.getUnprocessedSwitches()

    for switch in switches:
        opt = switch[0].lower()

        if opt in ("t", "start"):
            start = True
        if opt in ("p", "stop"):
            stop = True

    if start and stop:
        print("ERROR: decide if you want to start or stop ( not both ).")
        DIRAC.exit(2)
    elif not (start or stop):
        print("ERROR: decide if you want to start or stop.")
        DIRAC.exit(2)
    elif start:
        action = "start"
    elif stop:
        action = "stop"

    for prodID in args:
        result = diracProd.production(prodID, action, disableCheck=False)
        if "Message" in result:
            errorList.append((prodID, result["Message"]))
            exitCode = 2
        elif not result:
            errorList.append((prodID, "Null result for production() call"))
            exitCode = 2
        else:
            exitCode = 0

    for error in errorList:
        print(f"ERROR {error}")

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set a (set of) LFNs as problematic in the FC and in the BK and
transformation system if all replicas are problematic."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    dmScript.registerSiteSwitches()

    Script.registerSwitch("", "Reset", "   Reset files to OK")
    Script.registerSwitch("", "Full", "   Give full list of files")
    Script.registerSwitch("", "NoAction", "   No action taken, just give stats")

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] [<LFN>] [<LFN>...]",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=False)

    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeSetProblematicFiles
    from DIRAC import exit

    exit(executeSetProblematicFiles(dmScript))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""UploadLogFile module is used to upload the files present in the working
directory."""
import os
import shutil
import glob
import random
import stat
import shlex

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.DataManagementSystem.Client.FailoverTransfer import FailoverTransfer
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult

from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Workflow.Modules.ModulesUtilities import zipFiles
from LHCbDIRAC.Core.Utilities.ProductionData import getLogPath
from LHCbDIRAC.Core.Utilities.ResolveSE import getDestinationSEList


class UploadLogFile(ModuleBase):
    """Upload to LogSE."""

    #############################################################################

    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""

        self.log = gLogger.getSubLogger("UploadLogFile")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.logSE = self.opsH.getValue("LogStorage/LogSE", "LogSE")
        self.logSizeLimit = self.opsH.getValue("LogFiles/SizeLimit", 1 * 1024 * 1024)
        self.logExtensions = self.opsH.getValue("LogFiles/Extensions", [])
        self.logFilePath = ""
        self.logLFNPath = ""
        self.logdir = ""
        self.failoverTransfer = None
        self.failoverSEs = []

    ######################################################################

    def _resolveInputVariables(self):
        super()._resolveInputVariables()

        if "LogTargetPath" in self.workflow_commons:
            self.logLFNPath = self.workflow_commons["LogTargetPath"]
        else:
            self.log.info("LogFilePath parameter not found, creating on the fly")
            result = getLogPath(self.workflow_commons, self.bkClient)
            if not result["OK"]:
                self.log.error("Could not create LogFilePath", result["Message"])
                return result
            self.logLFNPath = result["Value"]["LogTargetPath"][0]

        if not isinstance(self.logLFNPath, str):
            self.logLFNPath = self.logLFNPath[0]

    ######################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Main executon method."""

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            self._resolveInputVariables()

            self.request.RequestName = "job_%d_request.xml" % self.jobID
            self.request.JobID = self.jobID
            self.request.SourceComponent = "Job_%d" % self.jobID

            res = systemCall(0, shlex.split("ls -al"))
            if res["OK"] and res["Value"][0] == 0:
                self.log.info("The contents of the working directory...")
                self.log.info(str(res["Value"][1]))
            else:
                self.log.error("Failed to list the log directory", str(res["Value"][2]))

            self.log.debug(f"PRODUCTION_ID = {self.production_id}, JOB_ID = {self.prod_job_id} ")
            self.logdir = os.path.realpath(f"./job/log/{self.production_id}/{self.prod_job_id}")
            self.log.info("Selected log files will be temporarily stored", f"in {self.logdir}")

            ##########################################
            # First determine the files which should be saved
            self.log.info("Determining the files to be saved in the logs.")
            res = self._determineRelevantFiles()
            if not res["OK"]:
                self.log.error("Completely failed to select relevant log files.", res["Message"])
                return S_OK()
            selectedFiles = res["Value"]
            self.log.info("The following files were selected to be saved", "\n".join(selectedFiles))

            #########################################
            # Create a temporary directory containing these files
            self.log.info("Populating a temporary directory for selected files.")
            res = self.__populateLogDirectory(selectedFiles)
            if not res["OK"]:
                self.log.error("Completely failed to populate temporary log file directory.", res["Message"])
                self.setApplicationStatus("Failed To Populate Log Dir")
                return S_OK()
            self.log.debug(f"{self.logdir} populated with log files.")

            #########################################
            # Make sure all the files in the log directory have the correct permissions
            result = self.__setLogFilePermissions(self.logdir)
            if not result["OK"]:
                self.log.error("Could not set permissions of log files to 0755", f"with message:\n{result['Message']}")

            # zip all files
            zipFileName = os.path.basename(str(self.prod_job_id) + ".zip")
            try:
                res = zipFiles(zipFileName, selectedFiles, str(self.prod_job_id))
                if not res["OK"]:
                    self.log.error("Failed to create zip of log files", res["Message"])
                    self.setApplicationStatus("Failed to create zip of log files")
                    # We do not fail the job for this case
                    return S_OK()
            except OSError:
                self.log.error("Failed to create zip of log files", res["Message"])
                self.setApplicationStatus("Failed to create zip of log files")
                # We do not fail the job for this case
                return S_OK()

            # Instantiate the failover transfer client with the global request object
            if not self.failoverTransfer:
                self.failoverTransfer = FailoverTransfer(self.request)

            # Attempt to uplaod the zippped logs to the LogSE
            self.log.info("Transferring zipped log files", f"to the {self.logSE}")

            if not self._enableModule():
                self.log.info("Would have attempted to upload the zipped log files, but there's not JobID")
                return S_OK()

            # self.logFilePath is something like /lhcb/MC/2016/LOG/00095376/0000/
            # the zipFileName should have the same name, e.g. 00000381.zip
            zipPath = os.path.join(self.logFilePath, zipFileName)
            res = returnSingleResult(StorageElement(self.logSE).getURL(zipPath, protocol="https"))
            if not res["OK"]:
                self.log.warn("Could not get dynamic URL for log", res)
                # The rule for interpreting what is to be deflated can be found in /eos/lhcb/grid/prod/lhcb/logSE/.htaccess
                logHttpsURL = f"https://lhcb-dirac-logse.web.cern.ch/lhcb-dirac-logse/{zipPath}"
            else:
                logHttpsURL = res["Value"]

            self.log.info("putFile", f"{zipFileName} to {self.logSE}")

            res = returnSingleResult(StorageElement(self.logSE).putFile({zipPath: zipFileName}))
            if res["OK"]:
                self.log.info("Successfully upload log file", f"to {self.logSE}")
                self.log.info("Logs for this job may be retrieved", f"from {logHttpsURL}")
            else:
                self.log.error(
                    "Failed to upload log files", f"with message '{res['Message']}', now uploading to failover SE"
                )
                self._uploadLogToFailoverSE(zipFileName)

            # While it's the zip file that is uploaded, we set in job parameters its directory,
            # as the .zip is deflated automatically
            self.setJobParameter("Log URL", f"<a href=\"{logHttpsURL.replace('.zip','/')}\">Log file directory</a>")

            self.workflow_commons["Request"] = self.request

            return S_OK("Log Files uploaded")

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in UploadLogFile execute module", lException=e)
            return S_ERROR(str(e))

        finally:
            super().finalize()

    #############################################################################

    def _uploadLogToFailoverSE(self, zipFileName):
        """Recover the logs to a failover storage element."""

        # here because self.siteName is not known until execute() is invoked
        self.failoverSEs = getDestinationSEList("Tier1-Failover", self.siteName, outputmode="Any")
        random.shuffle(self.failoverSEs)
        self.log.info(
            "Attempting to store file", f"{zipFileName} to the following SE(s):\n{', '.join(self.failoverSEs)}"
        )

        fileDict = {zipFileName: {"lfn": self.logLFNPath, "workflowSE": self.failoverSEs}}
        metadata = self.getFileMetadata(fileDict)
        fileMetaDict = {
            "Size": metadata[zipFileName]["filedict"]["Size"],
            "LFN": metadata[zipFileName]["filedict"]["LFN"],
            "GUID": metadata[zipFileName]["filedict"]["GUID"],
            "Checksum": metadata[zipFileName]["filedict"]["Checksum"],
            "ChecksumType": metadata[zipFileName]["filedict"]["ChecksumType"],
        }

        result = self.failoverTransfer.transferAndRegisterFile(
            fileName=zipFileName,
            localPath=f"{os.getcwd()}/{zipFileName}",
            lfn=self.logLFNPath,
            destinationSEList=self.failoverSEs,
            fileMetaDict=fileMetaDict,
            masterCatalogOnly=True,
        )

        if not result["OK"]:
            self.log.error("Failed to upload logs to all failover destinations (the job will not fail for this reason")
            self.setApplicationStatus("Failed To Upload Logs")
        else:
            uploadedSE = result["Value"]["uploadedSE"]
            self.log.info("Uploaded logs to failover SE", uploadedSE)

            self.request = self.failoverTransfer.request

            self.__createLogUploadRequest(self.logSE, self.logLFNPath, uploadedSE)
            self.log.verbose("Successfully created failover request")

    def _determineRelevantFiles(self):
        """The files which are below a configurable size will be stored in the
        logs.

        This will typically pick up everything in the working directory
        minus the output data files.
        """
        logFileExtensions = [
            "*.txt",
            "*.log",
            "*.out",
            "*.output",
            "*.xml",
            "*.sh",
            "*.info",
            "*.err",
            # TODO: Remove once the prodConf migration is done
            "prodConf*.py",
            "prodConf*.json",
        ]  # '*.root',
        if self.logExtensions:
            self.log.debug(f"Using list of log extensions from CS:\n{', '.join(self.logExtensions)}")
            logFileExtensions = self.logExtensions
        else:
            self.log.debug(f"Using default list of log extensions:\n{', '.join(logFileExtensions)}")

        candidateFiles = []
        for ext in logFileExtensions:
            self.log.debug(f"Looking at log file wildcard: {ext}")
            globList = glob.glob(ext)
            for check in globList:
                if os.path.isfile(check):
                    self.log.debug(f"Found locally existing log file: {check}")
                    candidateFiles.append(check)

        return S_OK(candidateFiles)

    #############################################################################

    def __populateLogDirectory(self, selectedFiles):
        """A temporary directory is created for all the selected files.

        These files are then copied into this directory before being
        uploaded
        """
        # Create the temporary directory
        mkDir(self.logdir)
        # Set proper permissions
        self.log.info("Changing log directory permissions to 0755")
        try:
            os.chmod(self.logdir, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH + stat.S_IXOTH)
        except OSError as x:
            self.log.error("Could not set logdir permissions to 0755:", f"{self.logdir} ({str(x)})")
        # Populate the temporary directory
        try:
            for fileS in selectedFiles:
                destinationFile = f"{self.logdir}/{os.path.basename(fileS)}"
                shutil.copy(fileS, destinationFile)
        except shutil.Error:
            self.log.warn("src and dst are the same")
        except OSError as x:
            self.log.exception("Exception while trying to copy file.", str(fileS) + str(x))
            self.log.info("The following files will be skipped and can be considered lost", str(fileS))

        # Now verify the contents of our target log dir
        successfulFiles = os.listdir(self.logdir)
        if not successfulFiles:
            self.log.info("Failed to copy any files to the target directory.")
            return S_ERROR()

        self.log.debug(f"Prepared {self.logdir} files in the temporary directory.")
        return S_OK()

    #############################################################################

    def __createLogUploadRequest(self, targetSE, logFileLFN, uploadedSE):
        """Set a request to upload job log files from the output sandbox."""
        self.log.info("Setting log upload request", f"for {logFileLFN} at {targetSE}")

        logUpload = Operation()
        logUpload.Type = "LogUpload"
        logUpload.TargetSE = targetSE

        logFile = File()
        logFile.LFN = logFileLFN

        logUpload.addFile(logFile)
        self.request.addOperation(logUpload)

        logRemoval = Operation()
        logRemoval.Type = "RemoveFile"
        logRemoval.TargetSE = uploadedSE

        logRemoval.addFile(logFile)
        self.request.addOperation(logRemoval)

    #############################################################################

    def __setLogFilePermissions(self, logDir):
        """Sets the permissions of all the files in the log directory to ensure
        they are readable."""
        try:
            for toChange in os.listdir(logDir):
                if not os.path.islink(f"{logDir}/{toChange}"):
                    self.log.debug(f"Changing permissions of {logDir}/{toChange} to 0755")
                    os.chmod(
                        f"{logDir}/{toChange}",
                        stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH + stat.S_IXOTH,
                    )
        except OSError as x:
            self.log.error("Problem changing shared area permissions", str(x))
            return S_ERROR(x)

        return S_OK()


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#
